// ==============================================================
//                 ORBITER MODULE: hayabusa2
//                  
//
// hayabusa2.cpp
// Control module for hayabusa2 vessel class
//
// 
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"
#include "OrbiterSoundSDK40.h"
#include <cstdio>
#include <cmath>
#include <string>
#include <stdio.h>
#include <math.h>

// ==============================================================
// Some vessel parameters
// ==============================================================

const double PA_ISP = 2500;
const double PB_ISP = 27500;
const double PB_MAXRCSTH = 20;
const double PB_MAXRCSTH30 = 0.010;
const double ANTENNA_OPERATING_SPEED = 0.25;
const double ARRAY_OPERATING_SPEED = 0.05;

#define YELLOW RGB(255, 255, 80)
#define GREEN RGB(0, 225, 0)
char refname[256];

const VECTOR3 min1a_off = _V(0.393,0.555,-0.66);
const VECTOR3 min1b_off = _V(0.393,0.491,-0.66);
const VECTOR3 min2_off = _V(-0.393,0.523,-0.66);
const VECTOR3 mas_off = _V(0.402, 0, 0.36);
const VECTOR3 mar1_off = _V(0.3,0,-0.66);
const VECTOR3 mar2_off = _V(0.2103,0.2103,-0.66);
const VECTOR3 mar3_off = _V( 0, 0.3,-0.66);
const VECTOR3 mar4_off = _V(-0.2103,0.2103,-0.66);
const VECTOR3 mar5_off = _V(-0.3,0,-0.66);
const VECTOR3 imp_off = _V(0,0,-0.867);
const VECTOR3 src_off = _V(-0.106,0.9,-0.215);
const VECTOR3 dcam_off = _V(0,0.637,0.652);




// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
void VLiftCoeff (VESSEL *v, double aoa, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 9;
	static const double AOA[nabsc] = {-180*RAD,-60*RAD,-30*RAD, -2*RAD, 10*RAD,15*RAD,50*RAD,90*RAD,180*RAD};
	static const double CL[nabsc]  = {0,0,0,0,0,0,0,0,0};
	static const double CM[nabsc]  = {0,0,0,0,0,0,0,0,0};
	for (i = 0; i < nabsc-1 && AOA[i+1] < aoa; i++);
	double f = (aoa-AOA[i]) / (AOA[i+1]-AOA[i]);
	*cl = CL[i] + (CL[i+1]-CL[i]) * f;  // aoa-dependent lift coefficient
	*cm = CM[i] + (CM[i+1]-CM[i]) * f;  // aoa-dependent moment coefficient
	double saoa = sin(aoa);
	double pd = 0.03 + 0.4*saoa*saoa;  // profile drag
	*cd = pd + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
	// profile drag + (lift-)induced drag + transonic/supersonic wave (compressibility) drag
}

void HLiftCoeff (VESSEL *v, double beta, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 8;
	static const double BETA[nabsc] = {-180*RAD,-135*RAD,-90*RAD,-45*RAD,45*RAD,90*RAD,135*RAD,180*RAD};
	static const double CL[nabsc]   = {0,0,0,0,0,0,0,0};
	for (i = 0; i < nabsc-1 && BETA[i+1] < beta; i++);
	*cl = CL[i] + (CL[i+1]-CL[i]) * (beta-BETA[i]) / (BETA[i+1]-BETA[i]);
	*cm = 0.0;
	*cd = 0.03 + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
}

//double LiftCoeff (double aoa)
//{
//	const int nlift = 9;
//	static const double AOA[nlift] = {-180*RAD,-60*RAD,-30*RAD,-1*RAD,15*RAD,20*RAD,25*RAD,60*RAD,180*RAD};
//	static const double CL[nlift]  = {       0,      0,		0,		0,		0,		0,		0,     0,      0};
//	static const double SCL[nlift] = {(CL[1]-CL[0])/(AOA[1]-AOA[0]), (CL[2]-CL[1])/(AOA[2]-AOA[1]),
//		                              (CL[3]-CL[2])/(AOA[3]-AOA[2]), (CL[4]-CL[3])/(AOA[4]-AOA[3]),
//									  (CL[5]-CL[4])/(AOA[5]-AOA[4]), (CL[6]-CL[5])/(AOA[6]-AOA[5]),
//									  (CL[7]-CL[6])/(AOA[7]-AOA[6]), (CL[8]-CL[7])/(AOA[8]-AOA[7])};
//	int i;
//	for (i = 0; i < nlift-1 && AOA[i+1] < aoa; i++);
//
//	return CL[i] + (aoa-AOA[i])*SCL[i];
//}


class hayabusa2: public VESSEL3 {
public:
	
	hayabusa2 (OBJHANDLE hVessel, int flightmodel);
	~hayabusa2 ();

	void clbkSetClassCaps (FILEHANDLE cfg);
	void DefineAnimations (void);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	//void clbkPostStep (double simt, double simdt, double mjd);
	void clbkPostCreation();
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);
	virtual bool clbkLoadGenericCockpit ();
	void DoPro();
	void DoRetro();
	void ToggleEng();
	void DoKillRot();
	void JettisonSRC(void);
	void JettisonMascot(void);
	void JettisonImpactor(void);
	void JettisonDcam(void);
	void JettisonMinerva(void);
	void JettisonMarker(void);
	void PrintElements(void);
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
	int clbkConsumeDirectKey(char *kstate);
	int MyID;
	bool SelectTarget(char *name);
	THRUSTER_HANDLE th_rcs[12], th_group[4], th_main[3], th_lid;
	PROPELLANT_HANDLE rcspr, ion_rcspr;
	ANIMATIONCOMPONENT_HANDLE Section, Section2, Section3, Section4;
	OBJHANDLE MyObj;
	SURFHANDLE tex_ion, tex_rcs;  

private:
	UINT anim_ant, anim_fold;
	double ant_proc, fold_proc, xenon, lid_tmr, flash_tmr;
	int autoattsw, ant_status, fold_status, mas_status, min_status, mar_status, src_status, dcam_status, imp_status;
//int a;
int engsw;
int engch;
int display_sw, cam_sw, lid_sw, flash_sw;

SpotLight *spotlight1;
BEACONLIGHTSPEC spot_beacon;

VECTOR3 beaconpos, beaconcol;

};

hayabusa2::hayabusa2 (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
}

hayabusa2::~hayabusa2 ()
{
	//if (tex_ion) oapiDestroySurface(tex_ion);
	//if (tex_rcs) oapiDestroySurface(tex_rcs);
}

// Define animation sequences for moving parts
// --------------------------------------------------------------
void hayabusa2::DefineAnimations (void)
{
	// 1. Hoover
	static UINT HiGainAnt1Grp[2] = {0,1};
	static MGROUP_TRANSLATE HiGainAnt1 (0, HiGainAnt1Grp, 2, _V(0,0,-0.9));
	anim_ant = CreateAnimation (0);
	AddAnimationComponent (anim_ant, 0, 1, &HiGainAnt1);


	
	// 2. Solar arrays - folding
	anim_fold = CreateAnimation (0);
	static UINT ArrayLFoldGrp1[1] = {9};
	static UINT ArrayLFoldGrp2[1] = {10};
	static UINT ArrayLFoldGrp3[1] = {11};
	static UINT ArrayLFoldGrp4[1] = {12};



	static UINT ArrayRFoldGrp1[1] = {13};
	static UINT ArrayRFoldGrp2[1] = {14};
	static UINT ArrayRFoldGrp3[1] = {15};
	static UINT ArrayRFoldGrp4[1] = {16};



	static MGROUP_ROTATE ArrayLFold1 (0, ArrayLFoldGrp1, 1, _V(-0.558,0,0.582), _V(0,1,0), (float)(PI*0.5));
	Section=AddAnimationComponent (anim_fold, 0.5,   1, &ArrayLFold1);
	static MGROUP_ROTATE ArrayLFold2 (0, ArrayLFoldGrp2, 1, _V(-0.576,0,-0.665), _V(0,-1,0), (float)(PI));
	Section2=AddAnimationComponent (anim_fold, 0.5,   1, &ArrayLFold2, Section);
	static MGROUP_ROTATE ArrayLFold3 (0, ArrayLFoldGrp3, 1, _V(-0.592,-0.662,0), _V(0,0,1), (float)(PI));
	AddAnimationComponent (anim_fold, 0,   0.3, &ArrayLFold3, Section2);
	static MGROUP_ROTATE ArrayLFold4 (0, ArrayLFoldGrp4, 1, _V(-0.599,0.67,0), _V(0,0,-1), (float)(PI));
	AddAnimationComponent (anim_fold, 0.2,   0.5, &ArrayLFold4, Section2);
	

	static MGROUP_ROTATE ArrayRFold1 (0, ArrayRFoldGrp1, 1, _V(0.558,0,0.582), _V(0,-1,0), (float)(PI*0.5));
	Section3=AddAnimationComponent (anim_fold, 0.5,   1, &ArrayRFold1);
	static MGROUP_ROTATE ArrayRFold2 (0, ArrayRFoldGrp2, 1, _V(0.576,0,-0.665), _V(0,1,0), (float)(PI));
	Section4=AddAnimationComponent (anim_fold, 0.5,   1, &ArrayRFold2, Section3);
	static MGROUP_ROTATE ArrayRFold3 (0, ArrayRFoldGrp3, 1, _V(0.592,-0.662,0), _V(0,0,-1), (float)(PI));
	AddAnimationComponent (anim_fold, 0,   0.3, &ArrayRFold3, Section4);
	static MGROUP_ROTATE ArrayRFold4 (0, ArrayRFoldGrp4, 1, _V(0.599,0.67,0), _V(0,0,1), (float)(PI));
	AddAnimationComponent (anim_fold, 0.2,   0.5, &ArrayRFold4, Section4);


}



// ==============================================================
// hayabusa2 class interface
// ==============================================================

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void hayabusa2::clbkSetClassCaps (FILEHANDLE cfg)
{
	ant_proc = 0.0;
	ant_status = 0;
	fold_proc = 0.0;
	fold_status = 0;
	strcpy(refname,"Sun");
	xenon = 0;
	autoattsw = 0;
	engsw = 0;
	engch = 0;
	mas_status = 0;
	min_status = 0;
	mar_status = 0;
	src_status = 0;
	dcam_status = 0;
	imp_status = 0;
	display_sw = 0;
	cam_sw = 0;
	lid_sw = 0;
	lid_tmr = 0;
	flash_sw = 0;
	flash_tmr = 0;
	DefineAnimations ();
	// physical specs
	SetCW (1, 1, 1, 1);
	//SetWingAspect (0.7);
	//SetWingEffectiveness (2.5);
	SetRotDrag (_V(1,1,1));
	//if (GetFlightModel() >= 1) {
	//	SetPitchMomentScale (5e-4);
	//	SetBankMomentScale (5e-4);
	//}
	//SetTrimScale (0.05);
	SetPMI (_V(0.3,0.3,0.3));
	SetCrossSections (_V(3,4,3));
	SetSize (2);
	SetEmptyMass (490);
	//SetClipRadius(5);
	SetMeshVisibilityMode ((AddMesh ("hayabusa2/hayabusa2")), MESHVIS_EXTERNAL);
	SetMeshVisibilityMode ((AddMesh ("hayabusa2/src_stowed")), MESHVIS_EXTERNAL);
	SetMeshVisibilityMode ((AddMesh ("hayabusa2/mascot", &mas_off)), MESHVIS_EXTERNAL);
	SetMeshVisibilityMode ((AddMesh ("hayabusa2/minerva1", &min1a_off)), MESHVIS_EXTERNAL);
	SetMeshVisibilityMode ((AddMesh ("hayabusa2/minerva1", &min1b_off)), MESHVIS_EXTERNAL);
	SetMeshVisibilityMode ((AddMesh ("hayabusa2/impactor_stowed")), MESHVIS_EXTERNAL);
	SetMeshVisibilityMode ((AddMesh ("hayabusa2/dcam",&dcam_off)), MESHVIS_EXTERNAL);
	SetMeshVisibilityMode ((AddMesh ("hayabusa2/marker",&mar1_off)), MESHVIS_EXTERNAL);
	SetMeshVisibilityMode ((AddMesh ("hayabusa2/marker",&mar2_off)), MESHVIS_EXTERNAL);
	SetMeshVisibilityMode ((AddMesh ("hayabusa2/marker",&mar3_off)), MESHVIS_EXTERNAL);
	SetMeshVisibilityMode ((AddMesh ("hayabusa2/marker",&mar4_off)), MESHVIS_EXTERNAL);
	SetMeshVisibilityMode ((AddMesh ("hayabusa2/marker",&mar5_off)), MESHVIS_EXTERNAL);
	SetMeshVisibilityMode ((AddMesh ("hayabusa2/minerva2", &min2_off)), MESHVIS_EXTERNAL);
	SetCameraOffset (_V(0,0,0));
	//SetLiftCoeffFunc (LiftCoeff);
	SetTouchdownPoints (_V(0.1,0.8,-0.8), _V(0.512,-0.853,-0.8), _V(-0.512,-0.853,-0.8));
	SetSurfaceFrictionCoeff (1,1);
	SetDockParams (_V(0,0,-1), _V(0,0,-1), _V(0,1,0));

	// propellant resources
	const double PA_FUELMASS = 37;
	const double PB_FUELMASS = 73;
	double rcsmass = 37;
	double ion_rcsmass = 73;

	ion_rcspr = CreatePropellantResource (PB_FUELMASS, ion_rcsmass);

	rcspr = CreatePropellantResource (PA_FUELMASS, rcsmass);
	
	// ***************** thruster definitions *******************
	th_main[0] = CreateThruster (_V( 0,0,0), _V(0,1, 0), PB_MAXRCSTH30, ion_rcspr, PB_ISP);
	th_main[1] = CreateThruster (_V( 0,0,0), _V(0,1, 0), PB_MAXRCSTH30, ion_rcspr, PB_ISP);
	th_main[2] = CreateThruster (_V( 0,0,0), _V(0,1, 0), PB_MAXRCSTH30, ion_rcspr, PB_ISP);

	tex_ion = oapiRegisterExhaustTexture ("hayabusa2/ion_exhaust");
	AddExhaust (th_main[0], 2, 0.1, _V( 0.154,-0.95,-0.017), _V(0,-1, 0), tex_ion);
	AddExhaust (th_main[1], 2, 0.1, _V( -0.154,-0.95,-0.017), _V(0,-1, 0), tex_ion);
	AddExhaust (th_main[2], 2, 0.1, _V( 0.154,-0.95,-0.323), _V(0,-1, 0), tex_ion);
	
	
	
	
	
	
	th_rcs[ 0] = CreateThruster (_V( 0.512, 0.853, -0.75), _V(0,0,1), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 1] = CreateThruster (_V(-0.512, 0.853, -0.75), _V(0,0,1), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 2] = CreateThruster (_V( 0.512,-0.853, -0.75), _V(0,0,1), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 3] = CreateThruster (_V(-0.512,-0.853, -0.75), _V(0,0,1), PB_MAXRCSTH, rcspr, PA_ISP);
	
	th_rcs[ 4] = CreateThruster (_V( 0.512, 0.853, 0.75), _V(0,0,-1), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 5] = CreateThruster (_V(-0.512, 0.853, 0.75), _V(0,0,-1), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 6] = CreateThruster (_V( 0.512,-0.853, 0.75), _V(0,0,-1), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 7] = CreateThruster (_V(-0.512,-0.853, 0.75), _V(0,0,-1), PB_MAXRCSTH, rcspr, PA_ISP);
	
	th_rcs[ 8] = CreateThruster (_V( 0.544, 0.94, 0), _V(0,-1,0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[ 9] = CreateThruster (_V(-0.544, 0.94, 0), _V(0,-1,0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[10] = CreateThruster (_V( 0.544,-0.94, 0), _V(0,1,0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[11] = CreateThruster (_V(-0.544,-0.94, 0), _V(0,1,0), PB_MAXRCSTH, rcspr, PA_ISP);


	tex_rcs = oapiRegisterExhaustTexture ("exhaust_atrcs");
	for(int i=0;i<12;i++)
	{
	AddExhaust (th_rcs[i], 1, 0.03, (tex_rcs));
	}

	th_group[0] = th_rcs[ 2];
	th_group[1] = th_rcs[ 3];
	th_group[2] = th_rcs[ 4];
	th_group[3] = th_rcs[ 5];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[ 0];
	th_group[1] = th_rcs[ 1];
	th_group[2] = th_rcs[ 6];
	th_group[3] = th_rcs[ 7];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_PITCHDOWN);

	th_group[0] = th_rcs[ 9];
	th_group[1] = th_rcs[10];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKLEFT);

	th_group[0] = th_rcs[ 8];
	th_group[1] = th_rcs[11];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKRIGHT);

	th_group[0] = th_rcs[ 0];
	th_group[1] = th_rcs[ 2];
	th_group[2] = th_rcs[ 5];
	th_group[3] = th_rcs[ 7];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_YAWLEFT);

	th_group[0] = th_rcs[ 1];
	th_group[1] = th_rcs[ 3];
	th_group[2] = th_rcs[ 4];
	th_group[3] = th_rcs[ 6];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[ 0];
	th_group[1] = th_rcs[ 1];
	th_group[2] = th_rcs[ 2];
	th_group[3] = th_rcs[ 3];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_FORWARD);

	th_group[0] = th_rcs[ 4];
	th_group[1] = th_rcs[ 5];
	th_group[2] = th_rcs[ 6];
	th_group[3] = th_rcs[ 7];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_BACK);

	th_group[0] = th_rcs[10];
	th_group[1] = th_rcs[11];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_UP);

	th_group[0] = th_rcs[ 8];
	th_group[1] = th_rcs[ 9];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_DOWN);


	th_group[0] = th_main[ 0];
	th_group[1] = th_main[ 1];
	th_group[2] = th_main[ 2];
	CreateThrusterGroup (th_group, 3, THGROUP_MAIN);


	SURFHANDLE lidar = oapiRegisterExhaustTexture ("hayabusa2/lidar");
	th_lid	= CreateThruster (_V(0.353,0.727,-0.724), _V(0, 0, 1), 0, ion_rcspr, 1000000000);
	
	AddExhaust (th_lid, 40, 0.01,_V(0.353,0.727,-0.724),_V( 0.3535, 0.3535,-0.866), lidar);
	AddExhaust (th_lid, 40, 0.01,_V(0.353,0.727,-0.724),_V(-0.3535, 0.3535,-0.866), lidar);
	AddExhaust (th_lid, 40, 0.01,_V(0.353,0.727,-0.724),_V( 0.3535,-0.3535,-0.866), lidar);
	AddExhaust (th_lid, 40, 0.01,_V(0.353,0.727,-0.724),_V(-0.3535,-0.3535,-0.866), lidar);

	///////////////////////////////////////////////////////
	/////AirFoils


	CreateAirfoil3 (LIFT_VERTICAL, _V(0,0,10), VLiftCoeff, 0, 1, 2, 1);
	CreateAirfoil3 (LIFT_HORIZONTAL, _V(0,0,10), HLiftCoeff, 0, 1, 2, 1);

	/////////////  lights ////////////////////////

	beaconpos = _V(0.253,0.727,-0.824);
	beaconcol = _V(1,1,1);

	spot_beacon.shape = BEACONSHAPE_STAR;
	spot_beacon.pos = &beaconpos;
	spot_beacon.col = &beaconcol;
	spot_beacon.size = 0.5;
	spot_beacon.falloff = 0.3;
	spot_beacon.period = 0;
	spot_beacon.duration = 0;
	spot_beacon.tofs = 0;
	spot_beacon.active = false;
	AddBeacon (&spot_beacon);


	COLOUR4 col_d = {1,1,1,0};
	COLOUR4 col_s = {1,1,1,0};
	COLOUR4 col_a = {0,0,0,0};
	
	spotlight1 = (SpotLight*)AddSpotLight(_V(0.253,0.727,-0.724), _V(0,0,-1),200, 1, 0, 1e-3, 90*RAD, (150*RAD), col_d, col_s, col_a);
	spotlight1->Activate(false);

}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new hayabusa2 (hvessel, flightmodel);
}




// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (hayabusa2*)vessel;
}
// ==============================================================
// Overloaded clbkPostCreation callback functions
// ==============================================================

void hayabusa2::clbkPostCreation ()
{
	
	MyID=ConnectToOrbiterSoundDLL(GetHandle());
	ReplaceStockSound(MyID,"Sound\\_CustomVesselsSounds\\hayabusa2\\ion_eng.wav",	REPLACE_MAIN_THRUST);
	
	if (GetPropellantMass(rcspr)< 0.01 && GetPropellantMass(ion_rcspr)< 0.01 && oapiGetSimTime ()> 0.1)
	{
		double rcsprmass = 37;
		double ion_rcsprmass = 73;
		SetPropellantMass(rcspr, rcsprmass);
		SetPropellantMass(ion_rcspr, ion_rcsprmass);
	}
	else SetPropellantMass(ion_rcspr, xenon);

	MyObj = oapiGetGbodyByName(refname);

	if (ant_status == 2) 
	{
		SetTouchdownPoints (_V(0.1,0.8,-1.7), _V(0.512,-0.853,-0.8), _V(-0.512,-0.853,-0.8));
		VESSELSTATUS vs;
		GetStatus(vs);
		if (vs.status == 1)DefSetState(&vs);
	}
	///////////// mesh n mass control /////////////////////
	double em = 490;
	if (mar_status == 1)
	{
		DelMesh(7);
		em = em-0.1;
	}
	else if (mar_status == 2)
	{
		DelMesh(7);
		DelMesh(8);
		em = em-0.2;
	}
	else if (mar_status == 3)
	{
		DelMesh(7);
		DelMesh(8);
		DelMesh(9);
		em = em-0.3;
	}
	else if (mar_status == 4)
	{
		DelMesh(7);
		DelMesh(8);
		DelMesh(9);
		DelMesh(10);
		em = em-0.4;
	}
	else if (mar_status == 5)
	{
		DelMesh(7);
		DelMesh(8);
		DelMesh(9);
		DelMesh(10);
		DelMesh(11);
		em = em-0.5;
	}

	///////// minerva
	if (min_status == 1)
	{
		DelMesh(3);
		em = em-0.25;
	}
	else if (min_status == 2)
	{
		DelMesh(3);
		DelMesh(4);
		em = em-0.5;
	}
	else if (min_status == 3)
	{
		DelMesh(3);
		DelMesh(4);
		DelMesh(12);
		em = em-1;
	}

	if (mas_status == 1)
	{
		DelMesh(2);
		em = em-10;
	}

	if (imp_status == 1)
	{
		DelMesh(5);
		em = em-7;
	}

	if (dcam_status == 1)
	{
		DelMesh(6);
		em = em-0.1;
	}

	if (src_status == 1)
	{
		DelMesh(1);
		em = em-16.5;
	}
	SetEmptyMass (em);
}
void hayabusa2::clbkSaveState (FILEHANDLE scn)
{


	for (int i = 0; i < (int)strlen(refname); i++) 
	{
		if (refname[i] == ' ') 
		{
			refname[i] = '+';
			break;
		}
	}

	char cbuf[256];

	// default vessel parameters
	VESSEL3::clbkSaveState (scn);

	// custom parameters
	oapiWriteScenario_int (scn, "ENG", engsw);
	oapiWriteScenario_int (scn, "AATT", autoattsw);

	sprintf (cbuf, "%d %d %d %d %d %d", min_status, mar_status, mas_status, imp_status, src_status, dcam_status);
	oapiWriteScenario_string (scn, "STS", cbuf);

	sprintf (cbuf, "%s", refname);
	oapiWriteScenario_string (scn, "AREF", cbuf);
	sprintf (cbuf, "%d %0.4f", ant_status, ant_proc);
	oapiWriteScenario_string (scn, "SPR", cbuf);
	sprintf (cbuf, "%d %0.4f", fold_status, fold_proc);
	oapiWriteScenario_string (scn, "SOLP", cbuf);
	sprintf (cbuf, "%0.6f", GetPropellantMass(ion_rcspr));
	oapiWriteScenario_string (scn, "XENON", cbuf);
}
void hayabusa2::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;

	while (oapiReadScenario_nextline (scn, line)) {
		if (!strnicmp (line, "ENG", 3)) {
            sscanf (line+3, "%d", &engsw);
		}else if (!strnicmp (line, "AATT", 4)) {
            sscanf (line+4, "%d", &autoattsw);
		}else if (!strnicmp (line, "STS", 3)) {
			sscanf_s (line+3, "%d%d%d%d%d%d", &min_status, &mar_status, &mas_status, &imp_status, &src_status, &dcam_status);
		}else if (!strnicmp (line, "AREF", 4)) {
			sscanf_s (line+4, "%s", &refname);
		}else if (!strnicmp (line, "SPR", 3)) {
			sscanf (line+3, "%d%lf", &ant_status, &ant_proc);
		} else if (!strnicmp (line, "SOLP", 4)) {
			sscanf (line+4, "%d%lf", &fold_status, &fold_proc);
		} else if (!strnicmp (line, "XENON", 5)) {
			sscanf (line+5, "%lf", &xenon);
		}else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }
	SetAnimation (anim_ant, ant_proc);
	SetAnimation (anim_fold, fold_proc);
	SetPropellantMass(ion_rcspr, xenon);
	engch = 1;

	for (int i = 0; i < (int)strlen(refname); i++) 
	{
		if (refname[i] == '+') 
		{
			refname[i] = ' ';
			break;
		}
	}

	MyObj = oapiGetGbodyByName(refname);
	
}






int hayabusa2::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate) {

bool cbSelectTarget(void *id, char *str, void *data);

  if (!down) return 0; // only process keydown events

  if (KEYMOD_SHIFT (kstate)) 
  {
	  switch (key) 
	  {
		case OAPI_KEY_G: 
		if (mar_status < 5)JettisonMarker();
		return 1;

		case OAPI_KEY_K: 
		if (min_status < 3)JettisonMinerva();
		return 1;

		case OAPI_KEY_V: 
		if (mas_status == 0)JettisonMascot();
		return 1;

		case OAPI_KEY_B: 
		if (imp_status == 0)JettisonImpactor();
		return 1;

		case OAPI_KEY_N: 
		if (dcam_status == 0)JettisonDcam();
		return 1;

		case OAPI_KEY_M: 
		if (src_status == 0)JettisonSRC();
		return 1;

		case OAPI_KEY_NUMPAD5: 
		if  (autoattsw == 0) autoattsw = 3;
		else if (autoattsw == 1) autoattsw = 3;
		else if (autoattsw == 2) autoattsw = 3;
		else if (autoattsw == 3) autoattsw = 0;
		return 1;

		case OAPI_KEY_W: 
		PrintElements();
		return 1;
	  }
	  return true;
  } 
  else if (KEYMOD_CONTROL (kstate)) 
  {
  
  }
  else { // unmodified keys
    switch (key) {
	case OAPI_KEY_V: // prograde
		if  (autoattsw == 0) autoattsw = 1;
		else if (autoattsw == 1) autoattsw = 0;
		else if (autoattsw == 2) autoattsw = 1;
		else if (autoattsw == 3) autoattsw = 1;
		return 1;
	case OAPI_KEY_B: // retrograde
		if  (autoattsw == 0) autoattsw = 2;
		else if (autoattsw == 2) autoattsw = 0;
		else if (autoattsw == 1) autoattsw = 2;
		else if (autoattsw == 3) autoattsw = 2;
		return 1;
	case OAPI_KEY_K: // toggle engine
		if (engch == 0) engch = 1;
		if  (engsw == 0) engsw = 1;
		else if  (engsw == 1) engsw = 2;
		else if (engsw == 2) engsw = 0;
		return 1;
	case OAPI_KEY_G: // lock solp
		if (fold_status == 0) fold_status = 1;
		return 1;
	case OAPI_KEY_E: // view down
		if (cam_sw == 0) 
		{
			cam_sw = 1;
			SetCameraDefaultDirection (_V(0,0,-1));
			oapiCameraSetCockpitDir(0,0);
			oapiCameraAttach (GetHandle(), 2);
		}
		else
		{
			cam_sw = 0;
			SetCameraDefaultDirection (_V(0,0,1));
			oapiCameraSetCockpitDir(0,0);
			oapiCameraAttach (GetHandle(), 2);
		}
		return 1;
	case OAPI_KEY_W: // lidar
		if (lid_sw == 0) lid_sw = 1;
		else 
		{
			lid_sw = 0;
			lid_tmr = 0;
			SetThrusterLevel(th_lid, 0);
		}
		return 1;
	case OAPI_KEY_P: // hoover
		if (ant_status == 0) ant_status = 1;
		//else if (ant_status == 1) ant_status = 3;
		//else if (ant_status == 2) ant_status = 3;
		//else if (ant_status == 3) ant_status = 1;
		return 1;
	case OAPI_KEY_J: // lids
		if (display_sw == 0) display_sw = 1;
		else if (display_sw == 1) display_sw = 0;
		return 1;
	case OAPI_KEY_M: // flash
		if (flash_sw == 0) flash_sw = 1;
		else 
		{
			flash_sw = 0;
			flash_tmr = 0;
			spot_beacon.active = false;
			spotlight1->Activate(false);
		}
		return 1;
	case OAPI_KEY_N:
		{
			oapiOpenInputBox("Select Reference Object", cbSelectTarget, 0, 20, (void *)this);
		}
		return true;
	}
		 }
  
  return 0;
  }

bool cbSelectTarget(void *id, char *str, void *data)
{
	return (((hayabusa2 *)data)->SelectTarget(str));
}

bool hayabusa2::SelectTarget(char *Name)
{
	MyObj = oapiGetGbodyByName(Name);

	if (MyObj)
	{
		//if (ibsw = 1)
		//{
		strcpy(refname,Name);
		//ibsw = 0;
		//}
	return true;
	}
	else return false;

}

int hayabusa2::clbkConsumeDirectKey(char *kstate)
{
		VECTOR3 avel = _V(0,0,0);
		GetAngularVel(avel);
		double dt = oapiGetSimStep();
		double drot = RAD*2*dt;

if(autoattsw == 0)
{
if (KEYMOD_SHIFT (kstate))
{
if (KEYDOWN (kstate, OAPI_KEY_NUMPAD2))
		{
		if (avel.x < RAD*10)
			{
			avel.x = avel.x+drot;
			if (avel.x > RAD*10) avel.x = RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD8))
		{
		if (avel.x > -RAD*10)
			{
			avel.x = avel.x-drot;
			if (avel.x < -RAD*10) avel.x = -RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD1))
		{
		if (avel.y < RAD*10)
			{
			avel.y = avel.y+drot;
			if (avel.y > RAD*10) avel.y = RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD3))
		{
		if (avel.y > -RAD*10)
			{
			avel.y = avel.y-drot;
			if (avel.y < -RAD*10) avel.y = -RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD6))
		{
		if (avel.z < RAD*10)
			{			
			avel.z = avel.z+drot;
			if (avel.z > RAD*10) avel.z = RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD4))
		{
		if (avel.z > -RAD*10)
			{
			avel.z = avel.z-drot;
			if (avel.z < -RAD*10) avel.z = -RAD*10;
			}		
		}

		SetAngularVel(avel);
	}
}

return 0;
}

void hayabusa2::DoKillRot()
{
		VECTOR3 avel = _V(0,0,0);
		GetAngularVel(avel);
		double dt = oapiGetSimStep();
		double drot = RAD*2*dt;

		if (drot > abs(avel.x)) avel.x = 0;
		else avel.x = avel.x + (drot * ((abs(avel.x))/-avel.x));
		if (drot > abs(avel.y)) avel.y = 0;
		else avel.y = avel.y + (drot * ((abs(avel.y))/-avel.y));
		if (drot > abs(avel.z)) avel.z = 0;
		else avel.z = avel.z + (drot * ((abs(avel.z))/-avel.z));

		SetAngularVel (avel);
		if (avel.x == 0 && avel.y == 0 && avel.z == 0) autoattsw = 0;
}

void hayabusa2::DoPro()
{
	
	if (MyObj)
	
	{
	DeactivateNavmode (1);
	DeactivateNavmode (2);
	DeactivateNavmode (3);
	DeactivateNavmode (4);
	DeactivateNavmode (5);
	DeactivateNavmode (6);
	DeactivateNavmode (7);


	//char cbuf[256]="Sun";
	//OBJHANDLE hRef = oapiGetGbodyByName (cbuf);
	VECTOR3 earthpos;
	VECTOR3 earthvel;
	GetRelativePos (MyObj, earthpos);
	//GetGlobalPos (earthpos);
	GetRelativeVel (MyObj, earthvel);
	double poslength;
	double vellength;
	poslength = length(earthpos);
	vellength = length(earthvel);
	VECTOR3 pos;
	VECTOR3 vel;
	pos = _V(earthpos.x/poslength, earthpos.y/poslength, earthpos.z/poslength);
	vel = _V(earthvel.x/vellength, earthvel.y/vellength, earthvel.z/vellength);

	VECTOR3 yup;
	VECTOR3 yupn;
	VECTOR3 zforward;
	VECTOR3 xright;
	yupn = crossp(vel, pos);
	double yupnlength;
	yupnlength = length(yupn);
	yup = _V(yupn.x/yupnlength, yupn.y/yupnlength, yupn.z/yupnlength);
	xright = crossp (yup, vel);
	zforward = vel;

	VECTOR3 nuxright;
	VECTOR3 nuyup = zforward;
	VECTOR3 nuzforward = -yup;
	yup = nuyup;
	zforward = nuzforward;

	nuzforward = -xright;
	nuxright = zforward;
	xright = nuxright;
	zforward = nuzforward;

	double gamma = atan2( xright.x, yup.x);
	double beta = -1*(asin(zforward.x));
	double alpha = atan2(zforward.y, zforward.z);
	SetGlobalOrientation (_V(alpha, beta, (PI/2)-gamma));
	SetAngularVel (_V(0,0,0));
	}
	else autoattsw = 0;
}


void hayabusa2::DoRetro()
{
	if (MyObj)
	{
	DeactivateNavmode (1);
	DeactivateNavmode (2);
	DeactivateNavmode (3);
	DeactivateNavmode (4);
	DeactivateNavmode (5);
	DeactivateNavmode (6);
	DeactivateNavmode (7);

	VECTOR3 earthpos;
	VECTOR3 earthvel;
	GetRelativePos (MyObj, earthpos);
	GetRelativeVel (MyObj, earthvel);
	double poslength;
	double vellength;
	poslength = -length(earthpos);
	vellength = -length(earthvel);
	VECTOR3 pos;
	VECTOR3 vel;
	pos = _V(earthpos.x/poslength, earthpos.y/poslength, earthpos.z/poslength);
	vel = _V(earthvel.x/vellength, earthvel.y/vellength, earthvel.z/vellength);
	VECTOR3 yup;
	VECTOR3 yupn;
	VECTOR3 zforward;
	VECTOR3 xright;
	yupn = crossp(vel, pos);
	double yupnlength;
	yupnlength = length(yupn);
	yup = _V(yupn.x/yupnlength, yupn.y/yupnlength, yupn.z/yupnlength);
	xright = crossp (yup, vel);
	zforward = vel;

	VECTOR3 nuxright;
	VECTOR3 nuyup = zforward;
	VECTOR3 nuzforward = -yup;
	yup = nuyup;
	zforward = nuzforward;

	nuzforward = xright;
	nuxright = -zforward;
	xright = nuxright;
	zforward = nuzforward;

	double gamma = atan2( xright.x, yup.x);
	double beta = -1*(asin(zforward.x));
	double alpha = atan2(zforward.y, zforward.z);
	SetGlobalOrientation (_V(alpha, beta, (PI/2)-gamma));
	SetAngularVel (_V(0,0,0));
	}
	else autoattsw = 0;

}

bool hayabusa2::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{

	// draw the default HUD
	VESSEL3::clbkDrawHUD (mode, hps, skp);

	int s = hps->H;
	double d = (s*0.00130208);
	int sw = ((hps->W));
	int lw =  (int)(16*sw/1024);
	int lwoffset = sw - (18*lw);
	int hlw = (int) (lw/2);


	int roxl = 0;
	int royl = 0;

	double ds = s;
	double dsw = sw;
	double sc_ratio = ds/dsw;


	if (sc_ratio < 0.7284)
	{
	roxl = (lw*10);
	royl = (int) (-88*d);
	}



	int wd = (int)(152*d);
	int wc = (int)(168*d);
	int w0 = (int)(184*d);
	int w1 = (int) (200*d);
	int w2 = (int)(216*d);
	int w3 = (int)(232*d);
	int w4 = (int)(248*d);
	int w5 = (int)(264*d);
	int w6 = (int)(280*d);
	int w7 = (int)(296*d);
	int w8 = (int)(312*d);
	int w9 = (int)(328*d);
	int w10 = (int)(344*d);
	int w11 = (int)(360*d);
	//int w12 = (int)(376*d);
	//int w13 = (int)(392*d);
	//int w14 = (int)(408*d);
	//int w15 = (int)(424*d);

	//int t = (int) (225*d);
	//int t1 = (int) (285*d);
	//int t2 = (int) (180*d);


	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"J = Toggle Display:Controls/Payload");
	skp->Text ((10+roxl), (wc+royl), abuf,strlen(abuf));
	}

if (display_sw == 0)
{
	if (fold_status == 0)
	{
	
	char dbuf[256];
	sprintf(dbuf,"G = Deploy Solar Panels");
	skp->Text ((10+roxl), (w0+royl), dbuf,strlen(dbuf));
	}
	else if (ant_status == 0)
	{
	char dbuf[256];
	sprintf(dbuf,"P = Deploy Sample Collector");
	skp->Text ((10+roxl), (w0+royl), dbuf,strlen(dbuf));
	}



	if (autoattsw == 1)
	{
	skp->SetTextColor(0x0066FFFF);
	char gbuf[256];
	sprintf(gbuf,"V = Hold Prograde On/Off");
	skp->Text ((10+roxl), (w1+royl), gbuf,strlen(gbuf));
	}
	else
	{
	char gbuf[256];
	sprintf(gbuf,"V = Hold Prograde On/Off");
	skp->Text ((10+roxl), (w1+royl), gbuf,strlen(gbuf));
	}

	if (autoattsw == 2)
	{
	skp->SetTextColor(0x0066FFFF);
	char gbuf[256];
	sprintf(gbuf,"B = Hold Retrograde On/Off");
	skp->Text ((10+roxl), (w2+royl), gbuf,strlen(gbuf));
	}
	else
	{
	skp->SetTextColor(0x0066FF66);
	char gbuf[256];
	sprintf(gbuf,"B = Hold Retrograde On/Off");
	skp->Text ((10+roxl), (w2+royl), gbuf,strlen(gbuf));
	}
	
	if (MyObj)
	{
	skp->SetTextColor(0x0066FF66);
	char dbuf[256];
	sprintf(dbuf,"N = Select Reference Body : %s", (refname) );
	skp->Text ((10+roxl), (w3+royl), dbuf,strlen(dbuf));
	}
	else
	{
	skp->SetTextColor(0x0066FF66);
	char dbuf[256];
	sprintf(dbuf,"N = Select Reference Body : None");
	skp->Text ((10+roxl), (w3+royl), dbuf,strlen(dbuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char gbuf[256];
	sprintf(gbuf,"K = Toggle Engine x%d", (3-engsw));
	skp->Text ((10+roxl), (w4+royl), gbuf,strlen(gbuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char gbuf[256];
	sprintf(gbuf,"W = Lidar On/Off");
	skp->Text ((10+roxl), (w5+royl), gbuf,strlen(gbuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char gbuf[256];
	sprintf(gbuf,"M = Strobe On/Off");
	skp->Text ((10+roxl), (w6+royl), gbuf,strlen(gbuf));
	}

	{
	char hbuf[256];
	sprintf(hbuf,"Shift+NumPad = Reaction Wheel Att.Control");
	skp->Text ((10+roxl), (w7+royl), hbuf,strlen(hbuf));
	}

	if (autoattsw == 3)
	{
	skp->SetTextColor(0x0066FFFF);
	char gbuf[256];
	sprintf(gbuf,"                                          Kill Rot.Active");
	skp->Text ((10+roxl), (w7+royl), gbuf,strlen(gbuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char kbuf[256];
	sprintf(kbuf,"E = Set View Forward/Back");
	skp->Text ((10+roxl), (w8+royl), kbuf,strlen(kbuf));
	}


	{
	double m1 = GetEmptyMass();
	double m2 = GetPropellantMass(rcspr);
	double m3 = GetPropellantMass(ion_rcspr);
	double test = ((m1+m2+m3)/(m1+m2));
	double dv = (27500*(log(test)));
	char cbuf[256];
	sprintf(cbuf,"Xenon %0.3f kg : dV %0.2f m/s", m3, dv);
	skp->Text ((10+roxl), (w9+royl), cbuf,strlen(cbuf));
	}
	{
	double m1 = GetEmptyMass();
	double m2 = GetPropellantMass(rcspr);
	double m3 = GetPropellantMass(ion_rcspr);
	double test = ((m1+m2+m3)/(m1+m3));
	double dvr = (2500*(log(test)));
	char abuf[256];
	sprintf(abuf,"Hydrazine %0.3f kg : dV %0.2f m/s", m2, dvr);
	skp->Text ((10+roxl), (w10+royl), abuf,strlen(abuf));
	}
	

	{
	double mass = GetMass ();
	double throttle_level = GetThrusterGroupLevel(THGROUP_MAIN);
	double thmax = 0.01*(3-engsw);
	double accel = throttle_level*86400*(thmax/mass);
	char kbuf[256];
	sprintf(kbuf,"Ion Eng.Accel.= %0.3f m/s per day", accel);
	skp->Text ((10+roxl), (w11+royl), kbuf,strlen(kbuf));
	}

	
}

else
{
	if (mar_status < 5)
	{
	skp->SetTextColor(0x0066FF66);
	char gbuf[256];
	sprintf(gbuf,"Shift+G = Deploy Marker");
	skp->Text ((10+roxl), (w0+royl), gbuf,strlen(gbuf));
	}

	if (min_status < 3)
	{
	skp->SetTextColor(0x0066FF66);
	char gbuf[256];
	sprintf(gbuf,"Shift+K = Deploy Minerva");
	skp->Text ((10+roxl), (w1+royl), gbuf,strlen(gbuf));
	}

	if (mas_status < 1)
	{
	skp->SetTextColor(0x0066FF66);
	char gbuf[256];
	sprintf(gbuf,"Shift+V = Deploy Mascot");
	skp->Text ((10+roxl), (w2+royl), gbuf,strlen(gbuf));
	}

	if (imp_status < 1)
	{
	skp->SetTextColor(0x0066FF66);
	char gbuf[256];
	sprintf(gbuf,"Shift+B = Deploy Impactor");
	skp->Text ((10+roxl), (w3+royl), gbuf,strlen(gbuf));
	}

	if (dcam_status < 1)
	{
	skp->SetTextColor(0x0066FF66);
	char gbuf[256];
	sprintf(gbuf,"Shift+N = Deploy DCam3");
	skp->Text ((10+roxl), (w4+royl), gbuf,strlen(gbuf));
	}

	if (src_status < 1)
	{
	skp->SetTextColor(0x0066FF66);
	char gbuf[256];
	sprintf(gbuf,"Shift+M = Deploy SRC");
	skp->Text ((10+roxl), (w5+royl), gbuf,strlen(gbuf));
	}


}



	return true;
}

void hayabusa2::ToggleEng()
{
	double throttle_level = GetThrusterGroupLevel(THGROUP_MAIN);
	SetThrusterGroupLevel(THGROUP_MAIN, 0);
	if (engsw == 0) 
	{
		DelThrusterGroup(THGROUP_MAIN);
		th_group[0] = th_main[ 0];
		th_group[1] = th_main[ 1];
		th_group[2] = th_main[ 2];
		CreateThrusterGroup (th_group, 3, THGROUP_MAIN);
		engch = 0;
	}

	else if (engsw == 1) 
	{
		DelThrusterGroup(THGROUP_MAIN);
		th_group[0] = th_main[ 0];
		th_group[1] = th_main[ 1];
		CreateThrusterGroup (th_group, 2, THGROUP_MAIN);
		engch = 0;
	}

	else if (engsw == 2) 
	{
		DelThrusterGroup(THGROUP_MAIN);
		th_group[0] = th_main[ 0];
		CreateThrusterGroup (th_group, 1, THGROUP_MAIN);
		engch = 0;
	}

	SetThrusterGroupLevel(THGROUP_MAIN,throttle_level);
}



void hayabusa2::clbkPreStep(double SimT, double SimDT, double MJD) {
	//sprintf(oapiDebugString(), "ps fm %d", GetFlightModel());

	if(lid_sw == 1)
	{
		SetThrusterLevel(th_lid, 1);
		lid_tmr = lid_tmr+SimDT;
		if (lid_tmr > 0.05)
		{
			lid_sw = 2;
			lid_tmr = 0;
		}
	}
	else if(lid_sw == 2)
	{
		SetThrusterLevel(th_lid, 0);
		lid_tmr = lid_tmr+SimDT;
		if (lid_tmr > 0.05)
		{
			lid_sw = 1;
			lid_tmr = 0;
		}
	}
	else SetThrusterLevel(th_lid, 0);

	if (flash_sw == 1)
	{
		flash_tmr = 0;
		spot_beacon.active = true;
		spotlight1->Activate(true);
		flash_sw = 2;
	}
	else if (flash_sw == 2)
	{
		spot_beacon.active = true;
		spotlight1->Activate(true);
		flash_sw = 3;
	}
	else if (flash_sw == 3)
	{
		spot_beacon.active = true;
		spotlight1->Activate(true);
		flash_sw = 4;
	}
	else if (flash_sw == 4)
	{
		flash_tmr = flash_tmr+SimDT;
		spot_beacon.active = false;
		spotlight1->Activate(false);
		if (flash_tmr > 2) flash_sw = 1;
	}
	else 
	{
		spot_beacon.active = false;
		spotlight1->Activate(false);
	}

	//sprintf(oapiDebugString(), "%d", ibsw);

	if (autoattsw == 1) DoPro();
	if (autoattsw == 2) DoRetro();
	if (autoattsw == 3) DoKillRot();
	if (engch == 1) ToggleEng();

	/// animations

	// Animate hoover
	if (ant_status == 1) 
	{
		double da = SimDT * ANTENNA_OPERATING_SPEED;
		ant_proc = ant_proc + da;
		if (ant_proc > 1) 
		{
			ant_proc = 1;
			ant_status = 2;
			SetTouchdownPoints (_V(0.1,0.8,-1.6), _V(-0.512,-0.853,-0.8), _V(0.512,-0.853,-0.8));
		}
	}
	SetAnimation (anim_ant, ant_proc);

	///// solps
	if (fold_status == 1) 
	{
		double da = SimDT * ARRAY_OPERATING_SPEED;
		fold_proc = fold_proc + da;
		if (fold_proc > 1) 
		{
			fold_proc = 1;
			fold_status = 2;
		}
	}
	SetAnimation (anim_fold, fold_proc);
	
	
}
//void hayabusa2::clbkPostStep (double simt, double simdt, double mjd)
//{

//}

////////// JETTISONS //////////////////////////////

void hayabusa2::JettisonMarker()
{
	VECTOR3 ofs = _V(0,0,0);
	VECTOR3 vofs = _V(0,0,-0.1);
	char name[256];
	char classname[256];
	strcpy (classname, "hayabusa2/marker");

	if (mar_status == 0)
	{
		strcpy(name, "Marker1");
		ofs = mar1_off;
		DelMesh(7);
		mar_status = 1;
	}
	else if (mar_status == 1)
	{
		strcpy(name, "Marker2");
		ofs = mar2_off;
		DelMesh(8);
		mar_status = 2;
	}
	else if (mar_status == 2)
	{
		strcpy(name, "Marker3");
		ofs = mar3_off;
		DelMesh(9);
		mar_status = 3;
	}
	else if (mar_status == 3)
	{
		strcpy(name, "Marker4");
		ofs = mar4_off;
		DelMesh(10);
		mar_status = 4;
	}
	else if (mar_status == 4)
	{
		strcpy(name, "Marker5");
		ofs = mar5_off;
		DelMesh(11);
		mar_status = 5;
	}
  VESSELSTATUS vs;
  GetStatus(vs);
  VECTOR3 rvel;
  Local2Rel (ofs, vs.rpos);
  GlobalRot (vofs, rvel);
  vs.rvel = vs.rvel + rvel;
  vs.status = 0;
  oapiCreateVessel (name, classname, vs);

  double em = GetEmptyMass();
  SetEmptyMass((em-0.1));
}

void hayabusa2::JettisonMinerva()
{
	VECTOR3 ofs = _V(0,0,0);
	VECTOR3 vofs = _V(0,0,-0.1);
	char name[256];
	char classname[256];
	double em = GetEmptyMass();

	if (min_status == 0)
	{
		strcpy(name, "Minerva1A");
		strcpy (classname, "hayabusa2/minerva1");
		ofs = min1a_off;
		DelMesh(3);
		min_status = 1;
		SetEmptyMass((em-0.25));
	}
	else if (min_status == 1)
	{
		strcpy(name, "Minerva1B");
		strcpy (classname, "hayabusa2/minerva1");
		ofs = min1b_off;
		DelMesh(4);
		min_status = 2;
		SetEmptyMass((em-0.25));
	}
	else if (min_status == 2)
	{
		strcpy(name, "Minerva2");
		strcpy (classname, "hayabusa2/minerva2");
		ofs = min2_off;
		DelMesh(12);
		min_status = 3;
		SetEmptyMass((em-0.5));
	}


  VESSELSTATUS vs;
  GetStatus(vs);
  VECTOR3 rvel;
  Local2Rel (ofs, vs.rpos);
  GlobalRot (vofs, rvel);
  vs.rvel = vs.rvel + rvel;
  vs.status = 0;
  oapiCreateVessel (name, classname, vs);

}

void hayabusa2::JettisonMascot()
{
	VECTOR3 ofs = mas_off;
	VECTOR3 vofs = _V(0.0994,0, -0.0105);
	char name[256];
	char classname[256];
	strcpy (classname, "hayabusa2/mascot");
	strcpy(name, "Mascot");


  VESSELSTATUS vs;
  GetStatus(vs);
  VECTOR3 rvel;
  Local2Rel (ofs, vs.rpos);
  GlobalRot (vofs, rvel);
  vs.rvel = vs.rvel + rvel;
  vs.status = 0;
  oapiCreateVessel (name, classname, vs);

  DelMesh(2);
  mas_status = 1;
  double em = GetEmptyMass();
  SetEmptyMass((em-10));

}

void hayabusa2::JettisonImpactor()
{
	VECTOR3 ofs = imp_off;
	VECTOR3 vofs = _V(0,0,-0.1);
	char name[256];
	char classname[256];
	strcpy (classname, "hayabusa2/impactor");
	strcpy(name, "Impactor");

	VECTOR3 pos;
	VECTOR3 vel;
	GlobalRot (_V(1,0,0), pos);
	GlobalRot (_V(0,0,-1), vel);
	VECTOR3 yup;
	VECTOR3 yupn;
	VECTOR3 zforward;
	VECTOR3 xright;
	yupn = crossp(vel, pos);
	double yupnlength;
	yupnlength = length(yupn);
	yup = _V(yupn.x/yupnlength, yupn.y/yupnlength, yupn.z/yupnlength);
	xright = crossp (yup, vel);
	zforward = vel;
	
	double gamma = atan2( xright.x, yup.x);
	double beta = -1*(asin(zforward.x));
	double alpha = atan2(zforward.y, zforward.z);


  VESSELSTATUS vs;
  GetStatus(vs);
  VECTOR3 rvel;
  Local2Rel (ofs, vs.rpos);
  GlobalRot (vofs, rvel);
  vs.rvel = vs.rvel + rvel;
  vs.arot = _V(alpha, beta, (PI/2)-gamma);
  double va = -vs.vrot.z;
  double vb	= -vs.vrot.y;
  double vc = vs.vrot.x;
  vs.vrot.x = vc;
  vs.vrot.y = vb;
  vs.vrot.z = va+(PI*0.2);
  vs.status = 0;
  oapiCreateVessel (name, classname, vs);

  DelMesh(5);
  imp_status = 1;
  double em = GetEmptyMass();
  SetEmptyMass((em-7));
}

void hayabusa2::JettisonDcam()
{
	VECTOR3 ofs = dcam_off;
	VECTOR3 vofs = _V(0,0,0.1);
	char name[256];
	char classname[256];
	strcpy (classname, "hayabusa2/dcam");
	strcpy(name, "DCam3");


  VESSELSTATUS vs;
  GetStatus(vs);
  VECTOR3 rvel;
  Local2Rel (ofs, vs.rpos);
  GlobalRot (vofs, rvel);
  vs.rvel = vs.rvel + rvel;
  vs.status = 0;
  oapiCreateVessel (name, classname, vs);

  DelMesh(6);
  dcam_status = 1;
  double em = GetEmptyMass();
  SetEmptyMass((em-0.1));
}

void hayabusa2::JettisonSRC()
{
VECTOR3 ofs = src_off;
	VECTOR3 vofs = _V(0,0.3,0);
	char name[256];
	char classname[256];
	strcpy (classname, "hayabusa2/src");
	strcpy(name, "SRC");

	VECTOR3 pos;
	VECTOR3 vel;
	GlobalRot (_V(1,0,0), pos);
	GlobalRot (_V(0,1,0), vel);
	VECTOR3 yup;
	VECTOR3 yupn;
	VECTOR3 zforward;
	VECTOR3 xright;
	yupn = crossp(vel, pos);
	double yupnlength;
	yupnlength = length(yupn);
	yup = _V(yupn.x/yupnlength, yupn.y/yupnlength, yupn.z/yupnlength);
	xright = crossp (yup, vel);
	zforward = vel;
	
	double gamma = atan2( xright.x, yup.x);
	double beta = -1*(asin(zforward.x));
	double alpha = atan2(zforward.y, zforward.z);


  VESSELSTATUS vs;
  GetStatus(vs);
  VECTOR3 rvel;
  Local2Rel (ofs, vs.rpos);
  GlobalRot (vofs, rvel);
  vs.rvel = vs.rvel + rvel;
  vs.arot = _V(alpha, beta, (PI/2)-gamma);
  double va = vs.vrot.x;
  double vb	= vs.vrot.y;
  double vc = vs.vrot.z;
  vs.vrot.x = va;
  vs.vrot.y = -vc;
  vs.vrot.z = vb+(15*RAD);
  vs.status = 0;
  oapiCreateVessel (name, classname, vs);

  DelMesh(1);
  src_status = 1;
  double em = GetEmptyMass();
  SetEmptyMass((em-16.5));
}

bool hayabusa2::clbkLoadGenericCockpit ()
{
if(cam_sw == 1)
	{
	SetCameraDefaultDirection (_V(0,0,-1));
	oapiCameraSetCockpitDir(0,0);
	}
else
	{
	SetCameraDefaultDirection (_V(0,0,1));
	oapiCameraSetCockpitDir(0,0);
	}
return true;
}

void hayabusa2::PrintElements(void)
{

	double tmp = GetAtmTemperature();
	double pres = GetAtmPressure();
	double dens = GetAtmDensity();


////// make a date and time tag
 
	
	char sun[256];
	strcpy(sun, "Sun");
	OBJHANDLE hRef = oapiGetObjectByName(sun);
	ELEMENTS el;
	ORBITPARAM prm;
	GetElements (hRef,el,&prm);


	char sma_str[256];
	sprintf(sma_str, "%0.9f", (el.a/AU));

	char ecc_str[256];
	sprintf(ecc_str, "%0.9f", el.e);

	char inc_str[256];
	sprintf(inc_str, "%0.9f", (el.i*DEG));

	char lan_str[256];
	sprintf(lan_str, "%0.9f", (el.theta *DEG));

	double Agp = DEG*(el.omegab - el.theta);
	char agp_str[256];
	sprintf(agp_str, "%0.9f", Agp);

	char ma_str[256];
	sprintf(ma_str, "%0.9f", (prm.MnA *DEG));

	char mjd_str[256];
	sprintf(mjd_str, "%0.9f", oapiGetSimMJD());


	char filename[256] = "Doc\\Hayabusa2\\hayabusa2ltt.txt";
	FILEHANDLE hFile = oapiOpenFile(filename, FILE_APP, ROOT);
			if (!hFile)
			{
			hFile = oapiOpenFile(filename, FILE_OUT, ROOT);
			oapiWriteLine(hFile, sma_str);
			oapiCloseFile(hFile, FILE_OUT);
			oapiOpenFile(filename, FILE_APP, ROOT);
			oapiWriteLine(hFile, ecc_str);
			oapiWriteLine(hFile, inc_str);
			oapiWriteLine(hFile, lan_str);
			oapiWriteLine(hFile, agp_str);
			oapiWriteLine(hFile, ma_str);
			oapiWriteLine(hFile, mjd_str);
			oapiCloseFile(hFile, FILE_APP);
			}		

			else
			{
			oapiWriteLine(hFile, sma_str);
			oapiWriteLine(hFile, ecc_str);
			oapiWriteLine(hFile, inc_str);
			oapiWriteLine(hFile, lan_str);
			oapiWriteLine(hFile, agp_str);
			oapiWriteLine(hFile, ma_str);
			oapiWriteLine(hFile, mjd_str);
			oapiCloseFile(hFile, FILE_APP);
			}


}