// ==============================================================
//                 ORBITER MODULE: dcam
//
// dcam.cpp
// Control module for dcam vessel class
//
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"

// ==============================================================
// Some vessel parameters
// ==============================================================



// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
double LiftCoeff (double aoa)
{
	const int nlift = 9;
	static const double AOA[nlift] = {-180*RAD,-60*RAD,-30*RAD,-1*RAD,15*RAD,20*RAD,25*RAD,60*RAD,180*RAD};
	static const double CL[nlift]  = {       0,      0,		0,		0,		0,		0,		0,     0,      0};
	static const double SCL[nlift] = {(CL[1]-CL[0])/(AOA[1]-AOA[0]), (CL[2]-CL[1])/(AOA[2]-AOA[1]),
		                              (CL[3]-CL[2])/(AOA[3]-AOA[2]), (CL[4]-CL[3])/(AOA[4]-AOA[3]),
									  (CL[5]-CL[4])/(AOA[5]-AOA[4]), (CL[6]-CL[5])/(AOA[6]-AOA[5]),
									  (CL[7]-CL[6])/(AOA[7]-AOA[6]), (CL[8]-CL[7])/(AOA[8]-AOA[7])};
	int i;
	for (i = 0; i < nlift-1 && AOA[i+1] < aoa; i++);

	return CL[i] + (aoa-AOA[i])*SCL[i];
}

// ==============================================================
// dcam class interface
// ==============================================================

class dcam: public VESSEL2 {
public:
	dcam (OBJHANDLE hVessel, int flightmodel)
		: VESSEL2 (hVessel, flightmodel) {}
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	//void clbkPostCreation();
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
	virtual bool clbkLoadGenericCockpit ();
	void clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, HDC hDC);


private:

	int mfdsw;

};

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void dcam::clbkSetClassCaps (FILEHANDLE cfg)
{
	
	mfdsw = 0;
	// physical specs
	SetCW (0.3, 0.3, 0.6, 0.9);
	SetWingAspect (0.7);
	SetWingEffectiveness (2.5);
	SetRotDrag (_V(0.6,0.6,0.35));
	
	SetTrimScale (0.05);
	SetSize (1);
	SetEmptyMass (0.1);
	SetPMI (_V(1,1,1));
	SetCrossSections (_V(0.1,0.1,0.1));
	SetCameraOffset (_V(0,0.1,-0.4));
	SetCameraDefaultDirection (_V(0,1,0));
	SetCameraRotationRange ((PI*0.999999), (PI*0.999999), (PI*0.499999), (PI*0.499999));
	SetLiftCoeffFunc (LiftCoeff);
	SetTouchdownPoints (_V(0,1,-0.01), _V(1,0,-0.01), _V(-1,0,-0.01));
	SetMeshVisibilityMode ((AddMesh ("hayabusa2/dcam")), MESHVIS_ALWAYS);
	

	// propellant resources
	
	
	// ***************** thruster definitions *******************


}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new dcam (hvessel, flightmodel);
}
// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (dcam*)vessel;
}
// ==============================================================
// Overloaded clbkPostCreation callback functions
// ==============================================================


void dcam::clbkPreStep(double SimT, double SimDT, double MJD)
{
//if (mfdsw == 0)
//	{
	
//	oapiOpenMFD(MFD_NONE, MFD_LEFT);
//	oapiOpenMFD(MFD_NONE, MFD_RIGHT);
//	mfdsw = 1;
//	}
}


int dcam::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate) {
  if (!down) return 0; // only process keydown events

  if (KEYMOD_SHIFT (kstate)) 
  {

  } 
  else 
  { // unmodified keys
    switch (key) 
	{
	case OAPI_KEY_J: // boom
		char impname[8];
		strcpy(impname,"Impactor");
		OBJHANDLE hVes = oapiGetVesselByName (impname);
		if (hVes == NULL) return 1;
		else 
		{
		
			VESSEL *v = oapiGetVesselInterface (hVes);
			v->SendBufferedKey (OAPI_KEY_J);
			return 1;
		}
		//return 1;
	}
  }
  return 0;
}

bool dcam::clbkLoadGenericCockpit ()
{
SetCameraDefaultDirection (_V(0,1,0),PI);
//oapiSetDefNavDisplay(0);
//oapiSetHUDMode(HUD_DOCKING);

return true;
}


void dcam::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, HDC hDC)
{

	// draw the default HUD
	VESSEL2::clbkDrawHUD (mode, hps, hDC);
	
	//char cbuf[256];
	
	{
	char cbuf[256];
	sprintf(cbuf,"J = Detonate Impactor Explosive!");
	TextOut(hDC,10,170,cbuf,strlen(cbuf));
	}

}