// ==============================================================
//                 ORBITER MODULE: impactor
//
// impactor.cpp
// Control module for impactor vessel class
//
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"

// ==============================================================
// Some vessel parameters
// ==============================================================

const double ISP = 1;
const double MAXMAINTHS = 0;
const double MAXMAINTHQ = 0;
const double FUELMASS = 1;


// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
double LiftCoeff (double aoa)
{
	const int nlift = 9;
	static const double AOA[nlift] = {-180*RAD,-60*RAD,-30*RAD,-1*RAD,15*RAD,20*RAD,25*RAD,60*RAD,180*RAD};
	static const double CL[nlift]  = {       0,      0,		0,		0,		0,		0,		0,     0,      0};
	static const double SCL[nlift] = {(CL[1]-CL[0])/(AOA[1]-AOA[0]), (CL[2]-CL[1])/(AOA[2]-AOA[1]),
		                              (CL[3]-CL[2])/(AOA[3]-AOA[2]), (CL[4]-CL[3])/(AOA[4]-AOA[3]),
									  (CL[5]-CL[4])/(AOA[5]-AOA[4]), (CL[6]-CL[5])/(AOA[6]-AOA[5]),
									  (CL[7]-CL[6])/(AOA[7]-AOA[6]), (CL[8]-CL[7])/(AOA[8]-AOA[7])};
	int i;
	for (i = 0; i < nlift-1 && AOA[i+1] < aoa; i++);

	return CL[i] + (aoa-AOA[i])*SCL[i];
}

// ==============================================================
// impactor class interface
// ==============================================================

class impactor: public VESSEL2 {
public:
	impactor (OBJHANDLE hVessel, int flightmodel)
		: VESSEL2 (hVessel, flightmodel) {}
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	void clbkPostCreation();
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
	void clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, HDC hDC);
	THRUSTER_HANDLE th_charge[6],  th_impact;
	PROPELLANT_HANDLE hProp;

private:

	double timr;
	int impsts;
};

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void impactor::clbkSetClassCaps (FILEHANDLE cfg)
{
	impsts = 0;
	timr = 0;

	// physical specs
	SetCW (0.3, 0.3, 0.6, 0.9);
	SetWingAspect (0.7);
	SetWingEffectiveness (2.5);
	SetRotDrag (_V(0.6,0.6,0.35));
	
	SetTrimScale (0.05);
	SetSize (1);
	SetEmptyMass (7.0);
	SetPMI (_V(1,1,1));
	SetCrossSections (_V(0.1,0.1,0.1));
	SetCameraOffset (_V(0,0.8,-1));
	SetLiftCoeffFunc (LiftCoeff);
	SetTouchdownPoints (_V(0,1,-0.01), _V(1,0,-0.01), _V(-1,0,-0.01));
	SetVisibilityLimit (1e-5);
	SetMeshVisibilityMode ((AddMesh ("hayabusa2/impactor")), MESHVIS_ALWAYS);
	SetMeshVisibilityMode ((AddMesh ("hayabusa2/impactor_case")), MESHVIS_ALWAYS);
	

	// propellant resources
	
	hProp = CreatePropellantResource (FUELMASS, FUELMASS);
	// ***************** thruster definitions *******************

	PARTICLESTREAMSPEC exhaust_main = {
		0, 1, 5, 10, 0.7, 1, 20, 1.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 8,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1
	};

	SURFHANDLE exh = oapiRegisterExhaustTexture ("contrail2");

	PARTICLESTREAMSPEC exhaust_add = {
		0, 0.25, 3, 50, 1, 2, 0, 1.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1,
		exh};

	PARTICLESTREAMSPEC eximpact = {
		0, 2, 30, 10, 2, 7, 15, 1.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1
	};

	PARTICLESTREAMSPEC eximpact_add = {
		0, 0.5, 60, 50, 1, 3, 0, 1.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1,
		exh};

	//PARTICLESTREAMSPEC excharge = {
	//	0, 1, 60, 50, 3, 2, 100, 1.0, PARTICLESTREAMSPEC::EMISSIVE,
	//	PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
	//	PARTICLESTREAMSPEC::ATM_FLAT, 1, 1
	//};

	th_charge[0] = CreateThruster (_V(0,0,0), _V(0,0,1), MAXMAINTHQ, hProp, ISP);
	th_charge[1] = CreateThruster (_V(0,0,0), _V(0,0,-1), MAXMAINTHQ, hProp, ISP);
	th_charge[2] = CreateThruster (_V(0,0,0), _V(0,1,0), MAXMAINTHQ, hProp, ISP);
	th_charge[3] = CreateThruster (_V(0,0,0), _V(0,-1,0), MAXMAINTHQ, hProp, ISP);
	th_charge[4] = CreateThruster (_V(0,0,0), _V(1,0,0), MAXMAINTHQ, hProp, ISP);
	th_charge[5] = CreateThruster (_V(0,0,0), _V(-1,0,0), MAXMAINTHQ, hProp, ISP);
	th_impact = CreateThruster (_V(0,0,0), _V(0,0,-1), MAXMAINTHS, hProp, ISP);
	//CreateThrusterGroup (&th_main, 1, THGROUP_MAIN);

	//AddExhaustStream (th_charge, _V(0,0,0), &excharge);
	
	AddExhaustStream (th_impact, _V(0,0,0), &eximpact_add);
	AddExhaustStream (th_impact, _V(0,0,0), &eximpact);
	

	for (int i = 0; i < 6; i++) 
	{
	AddExhaustStream (th_charge[i], _V(0,0,0), &exhaust_main);
	AddExhaustStream (th_charge[i], _V(0,0,0), &exhaust_add);
	}
	

}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new impactor (hvessel, flightmodel);
}
// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (impactor*)vessel;
}
// ==============================================================
// Overloaded clbkPostCreation callback functions
// ==============================================================

void impactor::clbkPostCreation ()
{
	if (impsts == 2) DelMesh(1);
	else if (impsts > 2) 
	{
	DelMesh(1);
	DelMesh(0);
	SetMeshVisibilityMode ((AddMesh ("hayabusa2/impactor_crater")), MESHVIS_EXTERNAL);
	SetSize (10);
	}
}

void impactor::clbkSaveState (FILEHANDLE scn)
{
	//char cbuf[256];

	// default vessel parameters
	VESSEL2::clbkSaveState (scn);

	// custom parameters
	oapiWriteScenario_int (scn, "IMST", impsts);
	//oapiWriteScenario_int (scn, "AUTOATT", doprograde);
	//sprintf (cbuf, "%d %0.4f", ant_status, ant_proc);
	//oapiWriteScenario_string (scn, "SPR", cbuf);
	//sprintf (cbuf, "%d %0.4f", fold_status, fold_proc);
	//oapiWriteScenario_string (scn, "SOLP", cbuf);
}
void impactor::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;
	while (oapiReadScenario_nextline (scn, line)) {
		if (!strnicmp (line, "IMST", 4)) {
            sscanf (line+4, "%d", &impsts);
		}else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
		}
	}

}


int impactor::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate) {
  if (!down) return 0; // only process keydown events

  if (KEYMOD_SHIFT (kstate)) 
  {

  } 
  else 
  { // unmodified keys
    switch (key) 
	{
	case OAPI_KEY_J: // boom
		if  (impsts == 0) impsts = 1;
		return 1;
	}
  }
  return 0;
}


void impactor::clbkPreStep(double SimT, double SimDT, double MJD) {

	if (impsts == 1)
	{
		for (int i = 0; i < 6; i++) 
		{
		SetThrusterLevel(th_charge[i], 1);
		}
		timr = timr+SimDT;
		if (timr > 0.2)
		{
		VESSELSTATUS2 vs;
		memset (&vs, 0, sizeof(vs));
		vs.version = 2;
		GetStatusEx(&vs);
		VECTOR3 rval;
		VECTOR3 vofs = _V(0,0, 2000);
		GlobalRot (vofs, rval);
		vs.rvel.x = vs.rvel.x + rval.x;
		vs.rvel.y = vs.rvel.y + rval.y;
		vs.rvel.z = vs.rvel.z + rval.z;
		DefSetStateEx (&vs);

		for (int i = 0; i < 6; i++) 
		{
		SetThrusterLevel(th_charge[i], 0);
		}
		SetPropellantMass(hProp,1);
		DelMesh(1);
		timr = 0;
		impsts = 2;
		}
	}

	else if (impsts == 2)
	{
	
		if (GetAltitude() < 3)
		{

	// next block of code
	VESSELSTATUS2 vs;
	memset (&vs, 0, sizeof(vs));
	vs.version = 2;
	vs.rbody = GetSurfaceRef();
	vs.status = 1;
	OBJHANDLE hVessel = GetHandle();
	double longitude, latitude, radius;
	double heading;
	oapiGetEquPos(hVessel,&longitude,&latitude,&radius);
	oapiGetHeading(hVessel,&heading);
	vs.surf_lat = latitude;
	vs.surf_lng = longitude;
	vs.surf_hdg = heading;

	DefSetStateEx (&vs);
	DelMesh(0);
	SetMeshVisibilityMode ((AddMesh ("hayabusa2/impactor_crater")), MESHVIS_EXTERNAL);
	SetSize (10);
	SetThrusterLevel(th_impact, 1);

	//OBJHANDLE hFocus = oapiGetFocusObject();
	//oapiCameraAttach(hFocus, 2);

	double camap = oapiCameraAperture();
	oapiCameraSetAperture(camap);

	impsts = 3;

		}
	}

	else if (impsts == 3)
	{
		timr = timr+SimDT;
		if (timr > 0.3)
		{
		timr = 0;
		SetThrusterLevel(th_impact, 0);
		impsts = 4;
		}
	}
}

void impactor::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, HDC hDC)
{

	// draw the default HUD
	VESSEL2::clbkDrawHUD (mode, hps, hDC);

	//char cbuf[256];
	
	{
	char cbuf[256];
	sprintf(cbuf,"J = Detonate Impactor Explosive!");
	TextOut(hDC,10,170,cbuf,strlen(cbuf));
	}

}