// ==============================================================
//                 ORBITER MODULE: minerva1
//
// minerva1.cpp
// Control module for minerva1 vessel class
//
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"

// ==============================================================
// Some vessel parameters
// ==============================================================



// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
double LiftCoeff (double aoa)
{
	const int nlift = 9;
	static const double AOA[nlift] = {-180*RAD,-60*RAD,-30*RAD,-1*RAD,15*RAD,20*RAD,25*RAD,60*RAD,180*RAD};
	static const double CL[nlift]  = {       0,      0,		0,		0,		0,		0,		0,     0,      0};
	static const double SCL[nlift] = {(CL[1]-CL[0])/(AOA[1]-AOA[0]), (CL[2]-CL[1])/(AOA[2]-AOA[1]),
		                              (CL[3]-CL[2])/(AOA[3]-AOA[2]), (CL[4]-CL[3])/(AOA[4]-AOA[3]),
									  (CL[5]-CL[4])/(AOA[5]-AOA[4]), (CL[6]-CL[5])/(AOA[6]-AOA[5]),
									  (CL[7]-CL[6])/(AOA[7]-AOA[6]), (CL[8]-CL[7])/(AOA[8]-AOA[7])};
	int i;
	for (i = 0; i < nlift-1 && AOA[i+1] < aoa; i++);

	return CL[i] + (aoa-AOA[i])*SCL[i];
}

// ==============================================================
// minerva1 class interface
// ==============================================================

class minerva1: public VESSEL2 {
public:
	minerva1 (OBJHANDLE hVessel, int flightmodel)
		: VESSEL2 (hVessel, flightmodel) {}
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
	int clbkConsumeDirectKey(char *kstate);
	void clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, HDC hDC);

private:

	int impsts, rotsw;
};

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void minerva1::clbkSetClassCaps (FILEHANDLE cfg)
{
	impsts = 0;
	rotsw = 0;

	// physical specs
	SetCW (0.3, 0.3, 0.6, 0.9);
	SetWingAspect (0.7);
	SetWingEffectiveness (2.5);
	SetRotDrag (_V(0.6,0.6,0.35));
	
	SetTrimScale (0.05);
	SetSize (0.5);
	SetEmptyMass (0.25);
	SetPMI (_V(0.1,0.1,0.1));
	SetCrossSections (_V(0.1,0.1,0.1));
	SetCameraOffset (_V(0,0.8,-1));
	SetLiftCoeffFunc (LiftCoeff);
	SetTouchdownPoints (_V(0,-0.032,0.06), _V(-0.052,-0.032,-0.03), _V(0.052,-0.032,-0.03));
	SetVisibilityLimit (1e-5);
	SetMeshVisibilityMode ((AddMesh ("hayabusa2/minerva1")), MESHVIS_ALWAYS);

	

}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new minerva1 (hvessel, flightmodel);
}
// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (minerva1*)vessel;
}
// ==============================================================
// Overloaded clbkPostCreation callback functions
// ==============================================================





int minerva1::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate) {
  if (!down) return 0; // only process keydown events

  if (KEYMOD_SHIFT (kstate)) 
  {

  } 
  else 
  { // unmodified keys
    switch (key) 
	{
	case OAPI_KEY_J: // jump
		impsts = 2;
		return 1;
	}
  }
  return 0;
}

int minerva1::clbkConsumeDirectKey(char *kstate)
{
	if (KEYMOD_SHIFT (kstate))
	{
	}
	else if (KEYMOD_CONTROL (kstate))
	{
	}
	else if(rotsw == 0)
	{
		if (KEYDOWN (kstate, OAPI_KEY_B)) rotsw = 1;
		
		else if (KEYDOWN (kstate, OAPI_KEY_M)) rotsw = 2;
		
	}

return 0;
}

void minerva1::clbkPreStep(double SimT, double SimDT, double MJD) {

	if (impsts == 2)
	{
	VESSELSTATUS vs;
	GetStatus(vs);
	if (vs.status == 1)
		{
		VECTOR3 vofs = _V(0,0.02,0.02);
		VECTOR3 ofs = _V(0,0.03,0);
		VECTOR3 rvel;
		GlobalRot (vofs, rvel);
		Local2Rel (ofs, vs.rpos);
		vs.status = 0;
		vs.rvel = vs.rvel + rvel;
		vs.vrot.x = vs.vrot.x - (15*RAD);
		vs.vrot.y = vs.vrot.y - (5*RAD);
		vs.vrot.z = vs.vrot.z;
		DefSetState(&vs);
		impsts = 1;
		}
		else impsts = 0;

	}
	else if (impsts == 1)
	{
		if (GetAltitude() > 0.2) impsts = 0;
	}

	else if (GetAltitude() < 0.07 && impsts == 0)
	{
	VESSELSTATUS2 vs;
	memset (&vs, 0, sizeof(vs));
	vs.version = 2;
	vs.rbody = GetSurfaceRef();
	vs.status = 1;
	OBJHANDLE hVessel = GetHandle();
	double longitude, latitude, radius;
	double heading;
	oapiGetEquPos(hVessel,&longitude,&latitude,&radius);
	oapiGetHeading(hVessel,&heading);
	vs.surf_lat = latitude;
	vs.surf_lng = longitude;
	vs.surf_hdg = heading;
	if (rotsw == 1)
	{
	vs.surf_hdg = vs.surf_hdg - (10*RAD*SimDT);
	if (vs.surf_hdg > PI2) vs.surf_hdg = vs.surf_hdg - PI2;
	else if (vs.surf_hdg < 0) vs.surf_hdg = vs.surf_hdg + PI2;
	rotsw = 0;
	}
	else if (rotsw == 2)
	{
	vs.surf_hdg = vs.surf_hdg + (10*RAD*SimDT);
	if (vs.surf_hdg > PI2) vs.surf_hdg = vs.surf_hdg - PI2;
	else if (vs.surf_hdg < 0) vs.surf_hdg = vs.surf_hdg + PI2;
	rotsw = 0;
	}

	DefSetStateEx (&vs);
	//sprintf(oapiDebugString(), "killed");
	}




}


void minerva1::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, HDC hDC)
{

	// draw the default HUD
	VESSEL2::clbkDrawHUD (mode, hps, hDC);

	//char cbuf[256];
	
	{
	char cbuf[256];
	sprintf(cbuf,"J = Jump");
	TextOut(hDC,10,170,cbuf,strlen(cbuf));
	}

	{
	char cbuf[256];
	sprintf(cbuf,"B/M = Rotate Left/Right");
	TextOut(hDC,10,186,cbuf,strlen(cbuf));
	}

}