// ==============================================================
//                 ORBITER MODULE: src
//                  
//
// src.cpp
// Control module for src vessel class
//
// 
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"
//#include <cstdio>
//#include <cmath>
//#include <string>
#include <stdio.h>
#include <math.h>

// ==============================================================
// Some vessel parameters
// ==============================================================


const double CHUTE_OPERATING_SPEED = 0.1;
const double par_size = 25;
const double par_ref = 5;
const double cap_size = 2;
const double cap_ref = 0.5;



// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
void VLiftCoeff (VESSEL *v, double aoa, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 9;
	static const double AOA[nabsc] = {-180*RAD,-60*RAD,-30*RAD, -2*RAD, 10*RAD,15*RAD,50*RAD,90*RAD,180*RAD};
	static const double CL[nabsc]  = {0,0,0,0,0,0,0,0,0};
	static const double CM[nabsc]  = {0,0,0,0,0,0,0,0,0};
	for (i = 0; i < nabsc-1 && AOA[i+1] < aoa; i++);
	double f = (aoa-AOA[i]) / (AOA[i+1]-AOA[i]);
	*cl = CL[i] + (CL[i+1]-CL[i]) * f;  // aoa-dependent lift coefficient
	*cm = CM[i] + (CM[i+1]-CM[i]) * f;  // aoa-dependent moment coefficient
	double saoa = sin(aoa);
	double pd = 0.03 + 0.4*saoa*saoa;  // profile drag
	*cd = pd + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
	// profile drag + (lift-)induced drag + transonic/supersonic wave (compressibility) drag
}

void HLiftCoeff (VESSEL *v, double beta, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 8;
	static const double BETA[nabsc] = {-180*RAD,-135*RAD,-90*RAD,-45*RAD,45*RAD,90*RAD,135*RAD,180*RAD};
	static const double CL[nabsc]   = {0,0,0,0,0,0,0,0};
	for (i = 0; i < nabsc-1 && BETA[i+1] < beta; i++);
	*cl = CL[i] + (CL[i+1]-CL[i]) * (beta-BETA[i]) / (BETA[i+1]-BETA[i]);
	*cm = 0.0;
	*cd = 0.03 + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
}


class src: public VESSEL3 {
public:
	
	src (OBJHANDLE hVessel, int flightmodel);
	~src ();

	void clbkSetClassCaps (FILEHANDLE cfg);
	void DefineAnimations (void);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	void clbkPostStep (double simt, double simdt, double mjd);
	void clbkPostCreation();
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	void Jettison(void);
	

private:
	UINT anim_chute;
	double chute_proc;
	int chute_status;
	AIRFOILHANDLE hHA, hVA;

};

src::src (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
}

src::~src ()
{
	//if (tex_ion) oapiDestroySurface(tex_ion);
	//if (tex_rcs) oapiDestroySurface(tex_rcs);
}

// Define animation sequences for moving parts
// --------------------------------------------------------------
void src::DefineAnimations (void)
{
	// 1. chute reefed
	static UINT HiGainAnt1Grp[1] = {0};
	static MGROUP_SCALE HiGainAnt1 (1, HiGainAnt1Grp, 1, _V(0,0,0),_V(100,100,200));
	anim_chute = CreateAnimation (0);
	AddAnimationComponent (anim_chute, 0, 0.1, &HiGainAnt1);


	
	// 2. unreef
	static UINT HiGainAnt2Grp[1] = {0};
	static MGROUP_SCALE HiGainAnt2 (1, HiGainAnt2Grp, 1, _V(0,0,0),_V(2,2,1));
	AddAnimationComponent (anim_chute, 0.9, 1, &HiGainAnt2);


}



// ==============================================================
// src class interface
// ==============================================================

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void src::clbkSetClassCaps (FILEHANDLE cfg)
{
	chute_proc = 0.0;
	chute_status = 0;
	DefineAnimations ();
	// physical specs
	SetCW (0.3, 0.6, 0.4, 0.4);
	//SetWingAspect (0.7);
	//SetWingEffectiveness (2.5);
	SetRotDrag (_V(2,2,1));
	//if (GetFlightModel() >= 1) {
	//	SetPitchMomentScale (5e-4);
	//	SetBankMomentScale (5e-4);
	//}
	//SetTrimScale (0.05);
	SetPMI (_V(0.2,0.2,0.4));
	SetCrossSections (_V(0.1,0.1,0.2));
	SetSize (1);
	SetEmptyMass (16.5);
	//SetClipRadius(5);
	SetMeshVisibilityMode ((AddMesh ("hayabusa2/src")), MESHVIS_EXTERNAL);
	SetMeshVisibilityMode ((AddMesh ("hayabusa2/src_chutes")), MESHVIS_NEVER);
	
	SetCameraOffset (_V(0,0,0));
	//SetLiftCoeffFunc (LiftCoeff);
	SetTouchdownPoints (_V(0.1,0.8,-0.08), _V(-0.512,-0.853,-0.08), _V(0.512,-0.853,-0.08));
	SetSurfaceFrictionCoeff (1,1);
	

	///////////////////////////////////////////////////////
	/////AirFoils


	hVA = CreateAirfoil3 (LIFT_VERTICAL, _V(0,0,-cap_ref), VLiftCoeff, 0, cap_size, cap_size, 1);
	hHA = CreateAirfoil3 (LIFT_HORIZONTAL, _V(0,0,-cap_ref), HLiftCoeff, 0, cap_size, cap_size, 1);

	// reentry pstream
	PARTICLESTREAMSPEC rps = {
		0, 1.8, 40, 0, 0.001, 0.5, 4, 0.2, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 1, PARTICLESTREAMSPEC::ATM_PLOG, 7e7, 1e8
	};
	AddReentryStream (&rps);

	
}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new src (hvessel, flightmodel);
}




// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (src*)vessel;
}
// ==============================================================
// Overloaded clbkPostCreation callback functions
// ==============================================================

void src::clbkPostCreation ()
{

	if (chute_status > 0) 
	{
		SetTouchdownPoints (_V(0,-0.1,0.2), _V(-0.2,-0.15,-0.2), _V(0.2,-0.15,-0.2));
		DelMesh(0);
		SetMeshVisibilityMode(1, MESHVIS_EXTERNAL);
		EditAirfoil (hVA,0x09, _V(0,0,-(cap_ref+(par_ref*chute_proc))), 0, 0, (cap_size+(par_size*chute_proc)), 0);
		EditAirfoil (hHA,0x09, _V(0,0,-(cap_ref+(par_ref*chute_proc))), 0, 0, (cap_size+(par_size*chute_proc)), 0);
		SetSize(10);
		SetPMI (_V(4,4,2));
	}
}
void src::clbkSaveState (FILEHANDLE scn)
{

	char cbuf[256];

	// default vessel parameters
	VESSEL3::clbkSaveState (scn);

	// custom parameters
	
	sprintf (cbuf, "%d %0.4f", chute_status, chute_proc);
	oapiWriteScenario_string (scn, "CHT", cbuf);
}
void src::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;

	while (oapiReadScenario_nextline (scn, line)) {
		if (!strnicmp (line, "CHT", 3)) {
			sscanf (line+3, "%d%lf", &chute_status, &chute_proc);
		} else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }
	SetAnimation (anim_chute, chute_proc);	
	
}




void src::clbkPreStep(double SimT, double SimDT, double MJD) {

	//sprintf(oapiDebugString(), "%d", ibsw);
	if (chute_status == 0)
	{
	if (GetAltitude() < 10000)
		{	
		chute_status = 1;
		Jettison();
		DelMesh(0);
		SetMeshVisibilityMode(1, MESHVIS_EXTERNAL);
		SetSize(10);
		SetPMI (_V(4,4,2));
		}
	}

	// Animate chute
	else if (chute_status == 1) 
	{
		double da = SimDT * CHUTE_OPERATING_SPEED;
		chute_proc = chute_proc + da;

		EditAirfoil (hVA,0x09, _V(0,0,-(cap_ref+(par_ref*chute_proc))), 0, 0, (cap_size+(par_size*chute_proc)), 0);
		EditAirfoil (hHA,0x09, _V(0,0,-(cap_ref+(par_ref*chute_proc))), 0, 0, (cap_size+(par_size*chute_proc)), 0);


		if (chute_proc > 1) 
		{
			chute_proc = 1;
			chute_status = 2;
			SetTouchdownPoints (_V(0,-0.1,0.2), _V(-0.2,-0.15,-0.2), _V(0.2,-0.15,-0.2));
		EditAirfoil (hVA,0x09, _V(0,0,-(cap_ref+(par_ref*chute_proc))), 0, 0, (cap_size+(par_size*chute_proc)), 0);
		EditAirfoil (hHA,0x09, _V(0,0,-(cap_ref+(par_ref*chute_proc))), 0, 0, (cap_size+(par_size*chute_proc)), 0);
		}
	}
	SetAnimation (anim_chute, chute_proc);

	
	
}
void src::clbkPostStep (double simt, double simdt, double mjd){


}

////////// JETTISONS //////////////////////////////

void src::Jettison()
{
	VECTOR3 ofs = _V(0,0,0);
	VECTOR3 vofs = _V(5,0,-5);
	char name[256];
	char classname[256];
	strcpy (classname, "hayabusa2/src_backshell");
	strcpy (name, "SRC_Backshell");

	
  VESSELSTATUS vs;
  GetStatus(vs);
  VECTOR3 rvel;
  Local2Rel (ofs, vs.rpos);
  GlobalRot (vofs, rvel);
  vs.rvel = vs.rvel + rvel;
  vs.vrot.x = vs.vrot.x + (90*RAD);
  vs.status = 0;
  oapiCreateVessel (name, classname, vs);

	ofs = _V(0,0,0);
	vofs = _V(0,0,2);
	strcpy (classname, "hayabusa2/src_heatshield");
	strcpy (name, "SRC_Heatshield");

  GetStatus(vs);
  Local2Rel (ofs, vs.rpos);
  GlobalRot (vofs, rvel);
  vs.rvel = vs.rvel + rvel;
  vs.status = 0;
  oapiCreateVessel (name, classname, vs);

}

