// ==============================================================
//                 ORBITER MODULE: procyon
//                  
//
// procyon.cpp
// Control module for procyon vessel class
//
// 
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"
#include "OrbiterSoundSDK40.h"
//#include <cstdio>
//#include <cmath>
//#include <string>
#include <stdio.h>
#include <math.h>

// ==============================================================
// Some vessel parameters
// ==============================================================

const double RCS_ISP = 240;
const double ION_ISP = 10000;
const double MAXRCSTH = 0.025;
const double MAXIONTH = 0.0003;
const double SOLP_SPEED = 1;
const double FUELMASS = 2.5;

#define YELLOW RGB(255, 255, 80)
#define GREEN RGB(0, 225, 0)
char refname[256];





// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
void VLiftCoeff (VESSEL *v, double aoa, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 9;
	static const double AOA[nabsc] = {-180*RAD,-60*RAD,-30*RAD, -2*RAD, 10*RAD,15*RAD,50*RAD,90*RAD,180*RAD};
	static const double CL[nabsc]  = {0,0,0,0,0,0,0,0,0};
	static const double CM[nabsc]  = {0,0,0,0,0,0,0,0,0};
	for (i = 0; i < nabsc-1 && AOA[i+1] < aoa; i++);
	double f = (aoa-AOA[i]) / (AOA[i+1]-AOA[i]);
	*cl = CL[i] + (CL[i+1]-CL[i]) * f;  // aoa-dependent lift coefficient
	*cm = CM[i] + (CM[i+1]-CM[i]) * f;  // aoa-dependent moment coefficient
	double saoa = sin(aoa);
	double pd = 0.03 + 0.4*saoa*saoa;  // profile drag
	*cd = pd + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
	// profile drag + (lift-)induced drag + transonic/supersonic wave (compressibility) drag
}

void HLiftCoeff (VESSEL *v, double beta, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 8;
	static const double BETA[nabsc] = {-180*RAD,-135*RAD,-90*RAD,-45*RAD,45*RAD,90*RAD,135*RAD,180*RAD};
	static const double CL[nabsc]   = {0,0,0,0,0,0,0,0};
	for (i = 0; i < nabsc-1 && BETA[i+1] < beta; i++);
	*cl = CL[i] + (CL[i+1]-CL[i]) * (beta-BETA[i]) / (BETA[i+1]-BETA[i]);
	*cm = 0.0;
	*cd = 0.03 + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
}


class procyon: public VESSEL3 {
public:
	
	procyon (OBJHANDLE hVessel, int flightmodel);
	~procyon ();

	void clbkSetClassCaps (FILEHANDLE cfg);
	void DefineAnimations (void);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	void clbkPostStep (double simt, double simdt, double mjd);
	void clbkPostCreation();
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);
	virtual bool clbkLoadGenericCockpit ();
	void DoPro();
	void DoRetro();
	void ToggleEng();
	void DoKillRot();
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
	int clbkConsumeDirectKey(char *kstate);
	int MyID;
	bool SelectTarget(char *name);
	THRUSTER_HANDLE th_rcs[12], th_group[4], th_main[1];
	PROPELLANT_HANDLE hProp;
	OBJHANDLE MyObj;
	SURFHANDLE tex_ion, tex_rcs;  

private:
	UINT anim_sp;
	double sp_proc, xenon;
	int autoattsw, sp_status;
	int cam_sw;


};

procyon::procyon (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
}

procyon::~procyon ()
{
	
}

// Define animation sequences for moving parts
// --------------------------------------------------------------
void procyon::DefineAnimations (void)
{
	// 1. solpanels
	anim_sp = CreateAnimation (0);

	static UINT HiGainAnt1Grp[1] = {2};
	static MGROUP_ROTATE HiGainAnt1 (0, HiGainAnt1Grp, 1, _V(0,0.278,0.311), _V(1,0,0), (90*RAD));
	AddAnimationComponent (anim_sp, 0, 1, &HiGainAnt1);

	static UINT HiGainAnt2Grp[1] = {3};
	static MGROUP_ROTATE HiGainAnt2 (0, HiGainAnt2Grp, 1, _V(0,-0.278,0.311), _V(-1,0,0), (90*RAD));
	AddAnimationComponent (anim_sp, 0, 1, &HiGainAnt2);

	static UINT HiGainAnt3Grp[1] = {5};
	static MGROUP_ROTATE HiGainAnt3 (0, HiGainAnt3Grp, 1, _V(0.278,0,0.311), _V(0,-1,0), (90*RAD));
	AddAnimationComponent (anim_sp, 0, 1, &HiGainAnt3);

	static UINT HiGainAnt4Grp[1] = {4};
	static MGROUP_ROTATE HiGainAnt4 (0, HiGainAnt4Grp, 1, _V(-0.278,0,0.311), _V(0, 1,0), (90*RAD));
	AddAnimationComponent (anim_sp, 0, 1, &HiGainAnt4);


	
	
}



// ==============================================================
// procyon class interface
// ==============================================================

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void procyon::clbkSetClassCaps (FILEHANDLE cfg)
{
	
	strcpy(refname,"Sun");
	xenon = 0;
	autoattsw = 0;
	cam_sw = 0;
	DefineAnimations ();
	// physical specs
	SetCW (1, 1, 1, 1);
	//SetWingAspect (0.7);
	//SetWingEffectiveness (2.5);
	SetRotDrag (_V(1,1,1));
	//if (GetFlightModel() >= 1) {
	//	SetPitchMomentScale (5e-4);
	//	SetBankMomentScale (5e-4);
	//}
	//SetTrimScale (0.05);
	SetPMI (_V(0.3,0.3,0.3));
	SetCrossSections (_V(0.3,0.3,0.3));
	SetSize (1);
	SetEmptyMass (64.5);
	//SetClipRadius(5);
	SetMeshVisibilityMode ((AddMesh ("hayabusa2/procyon")), MESHVIS_EXTERNAL);
	SetCameraOffset (_V(0,0,0));
	//SetLiftCoeffFunc (LiftCoeff);
	SetTouchdownPoints (_V(0.1,0.8,-0.3), _V(0.512,-0.853,-0.3), _V(-0.512,-0.853,-0.3));
	SetSurfaceFrictionCoeff (1,1);
	//SetDockParams (_V(0,0,-1), _V(0,0,-1), _V(0,1,0));

	// propellant resources
	


	hProp = CreatePropellantResource (FUELMASS, FUELMASS);

	
	// ***************** thruster definitions *******************
	th_main[0] = CreateThruster (_V( 0,0,0), _V(0,1,0), MAXIONTH, hProp, ION_ISP);


	tex_ion = oapiRegisterExhaustTexture ("hayabusa2/ion_exhaust");
	AddExhaust (th_main[0], 1, 0.04, _V(0, -0.32, 0), _V(0,-1, 0), tex_ion);
	

	
	th_rcs[ 0] = CreateThruster (_V(0, 0.282, 0.252), _V(0,-1,0), MAXRCSTH, hProp, RCS_ISP);
	th_rcs[ 1] = CreateThruster (_V(0, 0.282,-0.252), _V(0,-1,0), MAXRCSTH, hProp, RCS_ISP);
	th_rcs[ 2] = CreateThruster (_V(0,-0.282, 0.252), _V(0, 1,0), MAXRCSTH, hProp, RCS_ISP);
	th_rcs[ 3] = CreateThruster (_V(0,-0.282,-0.252), _V(0, 1,0), MAXRCSTH, hProp, RCS_ISP);
	
	th_rcs[ 4] = CreateThruster (_V(0.252, 0,-0.321), _V(0, 0, 1), MAXRCSTH, hProp, RCS_ISP);
	th_rcs[ 5] = CreateThruster (_V(-0.252, 0,-0.321), _V(0, 0, 1), MAXRCSTH, hProp, RCS_ISP);
	th_rcs[ 6] = CreateThruster (_V(0.252, 0, 0.321), _V(0, 0,-1), MAXRCSTH, hProp, RCS_ISP);
	th_rcs[ 7] = CreateThruster (_V(-0.252, 0, 0.321), _V(0, 0,-1), MAXRCSTH, hProp, RCS_ISP);
	
	th_rcs[ 8] = CreateThruster (_V(-0.28, 0.252, 0), _V( 1,0,0), MAXRCSTH, hProp, RCS_ISP);
	th_rcs[ 9] = CreateThruster (_V(-0.28,-0.252, 0), _V( 1,0,0), MAXRCSTH, hProp, RCS_ISP);
	th_rcs[10] = CreateThruster (_V(0.28, 0.252, 0), _V(-1,0,0), MAXRCSTH, hProp, RCS_ISP);
	th_rcs[11] = CreateThruster (_V(0.28,-0.252, 0), _V(-1,0,0), MAXRCSTH, hProp, RCS_ISP);


	tex_rcs = oapiRegisterExhaustTexture ("exhaust_atrcs");
	for(int i=0;i<12;i++)
	{
	AddExhaust (th_rcs[i], 0.5, 0.015, (tex_rcs));
	}

	

	

	th_group[0] = th_rcs[ 4];
	th_group[1] = th_rcs[ 5];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_FORWARD);

	th_group[0] = th_rcs[ 6];
	th_group[1] = th_rcs[ 7];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BACK);

	th_group[0] = th_rcs[ 2];
	th_group[1] = th_rcs[ 3];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_UP);

	th_group[0] = th_rcs[ 0];
	th_group[1] = th_rcs[ 1];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_DOWN);

	th_group[0] = th_rcs[10];
	th_group[1] = th_rcs[11];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_LEFT);

	th_group[0] = th_rcs[ 8];
	th_group[1] = th_rcs[ 9];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_RIGHT);


	th_group[0] = th_main[ 0];
	CreateThrusterGroup (th_group, 1, THGROUP_MAIN);



	///////////////////////////////////////////////////////
	/////AirFoils


	CreateAirfoil3 (LIFT_VERTICAL, _V(0,0,1), VLiftCoeff, 0, 0.1, 0.2, 1);
	CreateAirfoil3 (LIFT_HORIZONTAL, _V(0,0,1), HLiftCoeff, 0, 0.1, 0.2, 1);

}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new procyon (hvessel, flightmodel);
}




// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (procyon*)vessel;
}
// ==============================================================
// Overloaded clbkPostCreation callback functions
// ==============================================================

void procyon::clbkPostCreation ()
{
	MyID=ConnectToOrbiterSoundDLL(GetHandle());
	ReplaceStockSound(MyID,"Sound\\_CustomVesselsSounds\\hayabusa2\\ion_eng.wav",	REPLACE_MAIN_THRUST);
	
	if (GetPropellantMass(hProp)< 0.001 && oapiGetSimTime() > 0.1)
	{
		SetPropellantMass(hProp,FUELMASS);
	}
	else SetPropellantMass(hProp, xenon);

	MyObj = oapiGetGbodyByName(refname);

	
}
void procyon::clbkSaveState (FILEHANDLE scn)
{


	for (int i = 0; i < (int)strlen(refname); i++) 
	{
		if (refname[i] == ' ') 
		{
			refname[i] = '+';
			break;
		}
	}

	char cbuf[256];

	// default vessel parameters
	VESSEL3::clbkSaveState (scn);

	// custom parameters
	oapiWriteScenario_int (scn, "AATT", autoattsw);

	sprintf (cbuf, "%s", refname);
	oapiWriteScenario_string (scn, "AREF", cbuf);
	sprintf (cbuf, "%d %0.4f", sp_status, sp_proc);
	oapiWriteScenario_string (scn, "SPR", cbuf);
	sprintf (cbuf, "%lf", GetPropellantMass(hProp));
	oapiWriteScenario_string (scn, "XENON", cbuf);
}
void procyon::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;

	while (oapiReadScenario_nextline (scn, line)) {
		if (!strnicmp (line, "AATT", 4)) {
            sscanf (line+4, "%d", &autoattsw);
		}else if (!strnicmp (line, "AREF", 4)) {
			sscanf_s (line+4, "%s", &refname);
		}else if (!strnicmp (line, "SPR", 3)) {
			sscanf (line+3, "%d%lf", &sp_status, &sp_proc);
		} else if(!strnicmp (line, "XENON", 5)) {
			sscanf (line+5, "%lf", &xenon);
		}else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }
	SetAnimation (anim_sp, sp_proc);
	SetPropellantMass(hProp, xenon);

	for (int i = 0; i < (int)strlen(refname); i++) 
	{
		if (refname[i] == '+') 
		{
			refname[i] = ' ';
			break;
		}
	}

	MyObj = oapiGetGbodyByName(refname);
	
}






int procyon::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate) {

bool cbSelectTarget(void *id, char *str, void *data);

  if (!down) return 0; // only process keydown events

  if (KEYMOD_SHIFT (kstate)) 
  {
	  switch (key) 
	  {
		
		case OAPI_KEY_NUMPAD5: 
		if  (autoattsw == 0) autoattsw = 3;
		else if (autoattsw == 1) autoattsw = 3;
		else if (autoattsw == 2) autoattsw = 3;
		else if (autoattsw == 3) autoattsw = 0;
		return 1;
	  }
	  return true;
  } 
  else { // unmodified keys
    switch (key) {
	case OAPI_KEY_V: // prograde
		if  (autoattsw == 0) autoattsw = 1;
		else if (autoattsw == 1) autoattsw = 0;
		else if (autoattsw == 2) autoattsw = 1;
		else if (autoattsw == 3) autoattsw = 1;
		return 1;
	case OAPI_KEY_B: // retrograde
		if  (autoattsw == 0) autoattsw = 2;
		else if (autoattsw == 2) autoattsw = 0;
		else if (autoattsw == 1) autoattsw = 2;
		else if (autoattsw == 3) autoattsw = 2;
		return 1;
	
	case OAPI_KEY_G: // lock solp
		if (sp_status == 0) sp_status = 1;
		return 1;
	case OAPI_KEY_E: // view down
		if (cam_sw == 0) 
		{
			cam_sw = 1;
			SetCameraDefaultDirection (_V(0,0,-1));
			oapiCameraSetCockpitDir(0,0);
			oapiCameraAttach (GetHandle(), 2);
		}
		else
		{
			cam_sw = 0;
			SetCameraDefaultDirection (_V(0,0,1));
			oapiCameraSetCockpitDir(0,0);
			oapiCameraAttach (GetHandle(), 2);
		}
		return 1;
	
	case OAPI_KEY_N:
		{
			oapiOpenInputBox("Select Reference Object", cbSelectTarget, 0, 20, (void *)this);
		}
		return true;
	}
}
  
return 0;
}

bool cbSelectTarget(void *id, char *str, void *data)
{
	return (((procyon *)data)->SelectTarget(str));
}

bool procyon::SelectTarget(char *Name)
{
	MyObj = oapiGetGbodyByName(Name);

	if (MyObj)
	{
		//if (ibsw = 1)
		//{
		strcpy(refname,Name);
		//ibsw = 0;
		//}
	return true;
	}
	else return false;

}

int procyon::clbkConsumeDirectKey(char *kstate)
{
		VECTOR3 avel = _V(0,0,0);
		GetAngularVel(avel);
		double dt = oapiGetSimStep();
		double drot = RAD*2*dt;

if(autoattsw == 0)
{
if (KEYMOD_SHIFT (kstate))
{
if (KEYDOWN (kstate, OAPI_KEY_NUMPAD2))
		{
		if (avel.x < RAD*10)
			{
			avel.x = avel.x+drot;
			if (avel.x > RAD*10) avel.x = RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD8))
		{
		if (avel.x > -RAD*10)
			{
			avel.x = avel.x-drot;
			if (avel.x < -RAD*10) avel.x = -RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD1))
		{
		if (avel.y < RAD*10)
			{
			avel.y = avel.y+drot;
			if (avel.y > RAD*10) avel.y = RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD3))
		{
		if (avel.y > -RAD*10)
			{
			avel.y = avel.y-drot;
			if (avel.y < -RAD*10) avel.y = -RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD6))
		{
		if (avel.z < RAD*10)
			{			
			avel.z = avel.z+drot;
			if (avel.z > RAD*10) avel.z = RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD4))
		{
		if (avel.z > -RAD*10)
			{
			avel.z = avel.z-drot;
			if (avel.z < -RAD*10) avel.z = -RAD*10;
			}		
		}

		SetAngularVel(avel);
	}
}

return 0;
}

void procyon::DoKillRot()
{
		VECTOR3 avel = _V(0,0,0);
		GetAngularVel(avel);
		double dt = oapiGetSimStep();
		double drot = RAD*2*dt;

		if (drot > abs(avel.x)) avel.x = 0;
		else avel.x = avel.x + (drot * ((abs(avel.x))/-avel.x));
		if (drot > abs(avel.y)) avel.y = 0;
		else avel.y = avel.y + (drot * ((abs(avel.y))/-avel.y));
		if (drot > abs(avel.z)) avel.z = 0;
		else avel.z = avel.z + (drot * ((abs(avel.z))/-avel.z));

		SetAngularVel (avel);
		if (avel.x == 0 && avel.y == 0 && avel.z == 0) autoattsw = 0;
}

void procyon::DoPro()
{
	
	if (MyObj)
	
	{
	DeactivateNavmode (1);
	DeactivateNavmode (2);
	DeactivateNavmode (3);
	DeactivateNavmode (4);
	DeactivateNavmode (5);
	DeactivateNavmode (6);
	DeactivateNavmode (7);


	//char cbuf[256]="Sun";
	//OBJHANDLE hRef = oapiGetGbodyByName (cbuf);
	VECTOR3 earthpos;
	VECTOR3 earthvel;
	GetRelativePos (MyObj, earthpos);
	//GetGlobalPos (earthpos);
	GetRelativeVel (MyObj, earthvel);
	double poslength;
	double vellength;
	poslength = length(earthpos);
	vellength = length(earthvel);
	VECTOR3 pos;
	VECTOR3 vel;
	pos = _V(earthpos.x/poslength, earthpos.y/poslength, earthpos.z/poslength);
	vel = _V(earthvel.x/vellength, earthvel.y/vellength, earthvel.z/vellength);

	VECTOR3 yup;
	VECTOR3 yupn;
	VECTOR3 zforward;
	VECTOR3 xright;
	yupn = crossp(vel, pos);
	double yupnlength;
	yupnlength = length(yupn);
	yup = _V(yupn.x/yupnlength, yupn.y/yupnlength, yupn.z/yupnlength);
	xright = crossp (yup, vel);
	zforward = vel;

	VECTOR3 nuxright;
	VECTOR3 nuyup = zforward;
	VECTOR3 nuzforward = -yup;
	yup = nuyup;
	zforward = nuzforward;

	nuzforward = -xright;
	nuxright = zforward;
	xright = nuxright;
	zforward = nuzforward;

	double gamma = atan2( xright.x, yup.x);
	double beta = -1*(asin(zforward.x));
	double alpha = atan2(zforward.y, zforward.z);
	SetGlobalOrientation (_V(alpha, beta, (PI/2)-gamma));
	SetAngularVel (_V(0,0,0));
	}
	else autoattsw = 0;
}


void procyon::DoRetro()
{
	if (MyObj)
	{
	DeactivateNavmode (1);
	DeactivateNavmode (2);
	DeactivateNavmode (3);
	DeactivateNavmode (4);
	DeactivateNavmode (5);
	DeactivateNavmode (6);
	DeactivateNavmode (7);

	VECTOR3 earthpos;
	VECTOR3 earthvel;
	GetRelativePos (MyObj, earthpos);
	GetRelativeVel (MyObj, earthvel);
	double poslength;
	double vellength;
	poslength = -length(earthpos);
	vellength = -length(earthvel);
	VECTOR3 pos;
	VECTOR3 vel;
	pos = _V(earthpos.x/poslength, earthpos.y/poslength, earthpos.z/poslength);
	vel = _V(earthvel.x/vellength, earthvel.y/vellength, earthvel.z/vellength);
	VECTOR3 yup;
	VECTOR3 yupn;
	VECTOR3 zforward;
	VECTOR3 xright;
	yupn = crossp(vel, pos);
	double yupnlength;
	yupnlength = length(yupn);
	yup = _V(yupn.x/yupnlength, yupn.y/yupnlength, yupn.z/yupnlength);
	xright = crossp (yup, vel);
	zforward = vel;

	VECTOR3 nuxright;
	VECTOR3 nuyup = zforward;
	VECTOR3 nuzforward = -yup;
	yup = nuyup;
	zforward = nuzforward;

	nuzforward = xright;
	nuxright = -zforward;
	xright = nuxright;
	zforward = nuzforward;

	double gamma = atan2( xright.x, yup.x);
	double beta = -1*(asin(zforward.x));
	double alpha = atan2(zforward.y, zforward.z);
	SetGlobalOrientation (_V(alpha, beta, (PI/2)-gamma));
	SetAngularVel (_V(0,0,0));
	}
	else autoattsw = 0;

}

bool procyon::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{

	// draw the default HUD
	VESSEL3::clbkDrawHUD (mode, hps, skp);

	int s = hps->H;
	double d = (s*0.00130208);
	int sw = ((hps->W));
	int lw =  (int)(16*sw/1024);
	int lwoffset = sw - (18*lw);
	int hlw = (int) (lw/2);


	int roxl = 0;
	int royl = 0;

	double ds = s;
	double dsw = sw;
	double sc_ratio = ds/dsw;


	if (sc_ratio < 0.7284)
	{
	roxl = (lw*10);
	royl = (int) (-88*d);
	}



	int wd = (int)(152*d);
	int wc = (int)(168*d);
	int w0 = (int)(184*d);
	int w1 = (int) (200*d);
	int w2 = (int)(216*d);
	int w3 = (int)(232*d);
	int w4 = (int)(248*d);
	int w5 = (int)(264*d);
	int w6 = (int)(280*d);
	int w7 = (int)(296*d);
	int w8 = (int)(312*d);
	int w9 = (int)(328*d);
	int w10 = (int)(344*d);
	int w11 = (int)(360*d);
	//int w12 = (int)(376*d);
	//int w13 = (int)(392*d);
	//int w14 = (int)(408*d);
	//int w15 = (int)(424*d);

	//int t = (int) (225*d);
	//int t1 = (int) (285*d);
	//int t2 = (int) (180*d);


	


	if (sp_status == 0)
	{
	skp->SetTextColor(0x0066FF66);
	char dbuf[256];
	sprintf(dbuf,"G = Deploy Solar Panels");
	skp->Text ((10+roxl), (w0+royl), dbuf,strlen(dbuf));
	}
	else
	{
	char dbuf[256];
	sprintf(dbuf,"Solar panels Deployed");
	skp->Text ((10+roxl), (w0+royl), dbuf,strlen(dbuf));
	}



	if (autoattsw == 1)
	{
	skp->SetTextColor(0x0066FFFF);
	char gbuf[256];
	sprintf(gbuf,"V = Hold Prograde On/Off");
	skp->Text ((10+roxl), (w1+royl), gbuf,strlen(gbuf));
	}
	else
	{
	char gbuf[256];
	sprintf(gbuf,"V = Hold Prograde On/Off");
	skp->Text ((10+roxl), (w1+royl), gbuf,strlen(gbuf));
	}

	if (autoattsw == 2)
	{
	skp->SetTextColor(0x0066FFFF);
	char gbuf[256];
	sprintf(gbuf,"B = Hold Retrograde On/Off");
	skp->Text ((10+roxl), (w2+royl), gbuf,strlen(gbuf));
	}
	else
	{
	skp->SetTextColor(0x0066FF66);
	char gbuf[256];
	sprintf(gbuf,"B = Hold Retrograde On/Off");
	skp->Text ((10+roxl), (w2+royl), gbuf,strlen(gbuf));
	}
	
	if (MyObj)
	{
	skp->SetTextColor(0x0066FF66);
	char dbuf[256];
	sprintf(dbuf,"N = Select Reference Body : %s", (refname) );
	skp->Text ((10+roxl), (w3+royl), dbuf,strlen(dbuf));
	}
	else
	{
	skp->SetTextColor(0x0066FF66);
	char dbuf[256];
	sprintf(dbuf,"N = Select Reference Body : None");
	skp->Text ((10+roxl), (w3+royl), dbuf,strlen(dbuf));
	}

	{
	char hbuf[256];
	sprintf(hbuf,"Shift+NumPad = Reaction Wheel Att.Control");
	skp->Text ((10+roxl), (w4+royl), hbuf,strlen(hbuf));
	}

	if (autoattsw == 3)
	{
	skp->SetTextColor(0x0066FFFF);
	char gbuf[256];
	sprintf(gbuf,"                                          Kill Rot.Active");
	skp->Text ((10+roxl), (w4+royl), gbuf,strlen(gbuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char kbuf[256];
	sprintf(kbuf,"E = Set View Forward/Back");
	skp->Text ((10+roxl), (w5+royl), kbuf,strlen(kbuf));
	}


	{
	double m1 = GetEmptyMass();
	double m2 = GetPropellantMass(hProp);
	double test = ((m1+m2)/(m1));
	double dvion = (10000*(log(test)));
	double dvrcs = (240*(log(test)));
	char cbuf[256];
	sprintf(cbuf,"Xenon %0.3f kg : ION dV %0.1f m/s : RCS dV %0.1f m/s", m2, dvion, dvrcs);
	skp->Text ((10+roxl), (w6+royl), cbuf,strlen(cbuf));
	}
	
	{
	double mass = GetMass ();
	double throttle_level = GetThrusterGroupLevel(THGROUP_MAIN);
	double thmax = MAXIONTH;
	double accel = throttle_level*86400*(thmax/mass);
	char kbuf[256];
	sprintf(kbuf,"Ion Eng.Accel.= %0.4f m/s per day", accel);
	skp->Text ((10+roxl), (w7+royl), kbuf,strlen(kbuf));
	}



	return true;
}



void procyon::clbkPreStep(double SimT, double SimDT, double MJD) {

//sprintf(oapiDebugString(), "%d", ibsw);

	if (autoattsw == 1) DoPro();
	if (autoattsw == 2) DoRetro();
	if (autoattsw == 3) DoKillRot();

	/// animations


	///// solps
	if (sp_status == 1) 
	{
		double da = SimDT * SOLP_SPEED;
		sp_proc = sp_proc + da;
		if (sp_proc > 1) 
		{
			sp_proc = 1;
			sp_status = 2;
		}
	}
	SetAnimation (anim_sp, sp_proc);
	
	
}
void procyon::clbkPostStep (double simt, double simdt, double mjd){


}


bool procyon::clbkLoadGenericCockpit ()
{
if(cam_sw == 1)
	{
	SetCameraDefaultDirection (_V(0,0,-1));
	oapiCameraSetCockpitDir(0,0);
	}
else
	{
	SetCameraDefaultDirection (_V(0,0,1));
	oapiCameraSetCockpitDir(0,0);
	}
return true;
}